import atexit
import contextlib
from collections import namedtuple, OrderedDict
from datetime import datetime
import fasteners
import json
import os
import sqlite3
from threading import RLock
import uuid

from jolt import config
from jolt import expires
from jolt import filesystem as fs
from jolt import influence
from jolt import log
from jolt import tools
from jolt import utils
from jolt import tasks
from jolt.options import JoltOptions
from jolt.error import raise_error, raise_error_if
from jolt.error import raise_task_error, raise_task_error_if
from jolt.expires import ArtifactEvictionStrategyRegister


DEFAULT_ARCHIVE_TYPE = ".tar.gz"


def locked(func):
    def _f(self, *args, **kwargs):
        with self.lock:
            return func(self, *args, **kwargs)
    return _f


class StorageProvider(object):
    def download(self, artifact, force=False):
        return False

    def download_enabled(self):
        return True

    def upload(self, artifact, force=False):
        return False

    def upload_enabled(self):
        return True

    def location(self, artifact):
        return ''  # URL

    def availability(self, artifacts):
        # Ensure artifacts is a list
        artifacts = utils.as_list(artifacts)

        present = set()
        missing = set()

        for artifact in artifacts:
            if self.location(artifact):
                present.add(artifact)
            else:
                missing.add(artifact)

        return list(present), list(missing)


class StorageProviderFactory(StorageProvider):
    def create(self):
        pass


def RegisterStorage(cls):
    ArtifactCache.storage_provider_factories.append(cls)


class ArtifactAttributeSet(object):
    def __init__(self):
        super(ArtifactAttributeSet, self).__setattr__("_attributes", {})

    def _get_attributes(self):
        return self._attributes

    def __getattr__(self, name):
        attributes = self._get_attributes()
        if name not in attributes:
            attributes[name] = self.create(name)
        return attributes[name]

    def __setattr__(self, name, value):
        attributes = self._get_attributes()
        if name not in attributes:
            attributes[name] = self.create(name)
        attributes[name].set_value(value)
        return attributes[name]

    def __dict__(self):
        return {key: str(value) for key, value in self.items()}

    def items(self):
        return self._get_attributes().items()

    def apply(self, task, artifact):
        for _, value in self.items():
            value.apply(task, artifact)

    def apply_deps(self, task, deps):
        pass

    def unapply(self, task, artifact):
        for _, value in self.items():
            value.unapply(task, artifact)

    def unapply_deps(self, task, deps):
        pass

    def visit(self, task, artifact, visitor):
        for _, value in self.items():
            value.visit(task, artifact, visitor)


class ArtifactAttributeSetRegistry(object):
    providers = []

    @staticmethod
    def create_all(artifact):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().create(artifact)

    @staticmethod
    def parse_all(artifact, content):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().parse(artifact, content)

    @staticmethod
    def format_all(artifact, content):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().format(artifact, content)

    @staticmethod
    def apply_all(task, artifact):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().apply(task, artifact)

    @staticmethod
    def apply_all_deps(task, deps):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().apply_deps(task, deps)

    @staticmethod
    def unapply_all(task, artifact):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().unapply(task, artifact)

    @staticmethod
    def unapply_all_deps(task, deps):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().unapply_deps(task, deps)

    @staticmethod
    def visit_all(task, artifact, visitor):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().visit(task, artifact, visitor)


def visit_artifact(task, artifact, visitor):
    ArtifactAttributeSetRegistry.visit_all(task, artifact, visitor)


class ArtifactAttributeSetProvider(object):
    @staticmethod
    def Register(cls):
        ArtifactAttributeSetRegistry.providers.append(cls)

    def create(self, artifact):
        raise NotImplementedError()

    def parse(self, artifact, content):
        raise NotImplementedError()

    def format(self, artifact, content):
        raise NotImplementedError()

    def apply(self, task, artifact):
        pass

    def apply_deps(self, task, deps):
        pass

    def unapply(self, task, artifact):
        pass

    def unapply_deps(self, task, deps):
        pass

    def visit(self, task, artifact, visitor):
        pass


class ArtifactAttribute(object):
    def __init__(self, name):
        self._name = name

    def get_name(self):
        return self._name

    def set_value(self, value, expand=True):
        raise NotImplementedError()

    def get_value(self):
        raise NotImplementedError()

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass

    def __str__(self):
        raise NotImplementedError()


class ArtifactStringAttribute(ArtifactAttribute):
    def __init__(self, artifact, name):
        self._artifact = artifact
        self._name = name
        self._value = None

    def get_name(self):
        return self._name

    def set_value(self, value, expand=True):
        self._value = self._artifact.tools.expand(str(value)) if expand else str(value)

    def get_value(self):
        return self._value

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass

    def __str__(self):
        return str(self._value)


class ArtifactListAttribute(ArtifactAttribute):
    def __init__(self, artifact, name):
        self._artifact = artifact
        self._name = name
        self._value = []

    def get_name(self):
        return self._name

    def set_value(self, value, expand=True):
        if type(value) is str:
            value = value.split(":")
        raise_error_if(type(value) is not list, "Illegal value assigned to artifact list attribute")
        self._value = self._artifact.tools.expand(value) if expand else value

    def get_value(self):
        return self._value

    def append(self, value):
        if type(value) is list:
            self._value.extend(self._artifact.tools.expand(value))
        else:
            self._value.append(self._artifact.tools.expand(value))

    def extend(self, value):
        raise_error_if(
            type(value) is not list,
            "Illegal type passed to {}.extend() - list expected".format(self._name))
        self._value.extend(self._artifact.tools.expand(value))

    def items(self):
        return list(self._value)

    def count(self):
        return len(self.items())

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass


class ArtifactFileAttribute(object):
    def __init__(self):
        self._files = []

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass

    def append(self, src, dst):
        self._files.append((fs.as_posix(src), fs.as_posix(dst)))

    def assign(self, files):
        self._files = files

    def items(self):
        return self._files


@ArtifactAttributeSetProvider.Register
class ArtifactFileAttributeProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "files", ArtifactFileAttribute())

    def parse(self, artifact, content):
        if "files" not in content:
            return
        artifact.files.assign([])
        for record in content["files"]:
            artifact.files.append(record["src"], record["dst"])

    def format(self, artifact, content):
        content["files"] = [{"src": src, "dst": dst} for src, dst in artifact.files.items()]

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass

    def visit(self, task, artifact, visitor):
        pass


def json_serializer(obj):
    if isinstance(obj, datetime):
        return dict(type="datetime", value=obj.strftime("%Y-%m-%d %H:%M:%S.%f"))


def json_deserializer(dct):
    if dct.get("type") == "datetime":
        return datetime.strptime(dct["value"], "%Y-%m-%d %H:%M:%S.%f")
    return dct


class Artifact(object):
    """
    An artifact is a collection of files and metadata produced by a task.

    Task implementors call artifact methods to collect files to be published.
    In addition to files, other metadata can be provided as well, such as
    variables that should be set in the environment of consumer tasks.

    """

    cxxinfo = {}
    """ Artifact C/C++ build metadata.

    A task can add compilation metadata to an artifact. Such metadata
    will be automatically applied when consumer compilation tasks
    are executed. A common use-case is to add preprocessor definitions,
    link libraries, etc. These string fields are supported:

    - ``asflags`` - assembler flags (string)
    - ``cflags`` - compiler flags (string)
    - ``cxxflags`` - compiler flags (string)
    - ``ldflags`` - linker flags (string)
    - ``libraries`` - libraries to link with (list, use append())
    - ``macros`` - preprocessor macros to set (list, use append())
    - ``sources`` - source files to compile (list, use append())

    Values appended to PATH-type metadata fields are relative to the artifact
    root. They will be automatically expanded to absolute paths. These
    PATH-type fields are supported:

    - ``incpaths`` - preprocessor include paths (list, use append())
    - ``libpaths`` - linker library search paths (list, use append())

    Example:

        .. code-block:: python

            def publish(self, artifact, tools):
                artifact.collect("*.h", "include/")
                artifact.cxxinfo.incpaths.append("include")
                artifact.cxxinfo.macros.append("PACKAGE_VERSION=1.0")
    """

    environ = {}
    """ Artifact environment variables.

    A task can add environment variables to an artifact. Such a
    variable will automatically be set in the environment when
    consumer tasks are executed. A common use-case is to add
    programs to the PATH.

    Values appended to PATH variables are relative to the artifact
    root. They will be automatically expanded to absolute paths.
    This applies to all variables with ``PATH`` in the name.

    Example:

        .. code-block:: python

            def publish(self, artifact, tools):
                artifact.environ.PATH.append("bin")
                artifact.environ.JAVA_HOME = artifact.final_path
    """

    paths = {}
    """ Artifact paths.

    A task can add paths to files and directories inside an artifact.
    Paths are relative to the root of the artifact when created, but
    are expanded to absolute paths when the artifact is consumed by
    a task.

    This is useful as an abstraction when directories or filenames
    have varying names.

    Example:

        .. code-block:: python

            def publish(self, artifact, tools):
                artifact.paths.file = "{date}.txt"

    The ``file`` path is then expanded to a full path for consumers:

        .. code-block:: python

            requires = ["dep"]

            def run(self, deps, tools):
                filedata = tools.read_file(deps["dep"].paths.file)

    """

    python = {}
    """ Artifact Python configuration.

    A task can add Python configuration to an artifact. Such configuration
    will automatically be set in the environment when consumer tasks are
    executed. A common use-case is to add Python modules to the PATH so that
    they can be easily imported by a consumer.

    Values appended to PATH-type variables are relative to the artifact
    root. They will be automatically expanded to absolute paths.

    Example:

        .. code-block:: python

            def publish(self, artifact, tools):
                artifact.python.PATH.append("my_module")
    """

    strings = {}
    """ Artifact strings.

    A task can add arbitrary string values to an artifact. Such a
    string will be available for consumer tasks to read.

    Example:

        .. code-block:: python

            def publish(self, artifact, tools):
                artifact.strings.version = "1.2"
    """

    def __init__(self, cache, node, name=None, identity=None, tools=None, session=False):
        self._cache = cache
        if identity:
            self._identity = identity
        else:
            self._identity = node.identity if not session else node.instance
        if name:
            self._identity = name + "@" + self._identity
        self._main = name == "main"
        self._name = name or "main"
        self._node = node
        self._session = session
        self._task = node.task if node else None
        self._tools = tools or self._node.tools
        self._path = cache._fs_get_artifact_path(self._identity, node.canonical_name if node else name)
        self._temp = cache._fs_get_artifact_tmppath(self._identity, node.canonical_name if node else name)
        self._archive = cache._fs_get_artifact_archivepath(self._identity, node.canonical_name if node else name)
        self._lock_path = cache._fs_get_artifact_lockpath(self._identity)
        ArtifactAttributeSetRegistry.create_all(self)
        self.reload()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        pass

    def __getattr__(self, name):
        raise_task_error(self._node, "Attempt to access invalid artifact attribute '{0}'", name)

    def _write_manifest(self):
        content = {}
        content["size"] = self._get_size()
        content["unpacked"] = self._unpacked
        content["uploadable"] = self._uploadable
        if self._node:
            content["task"] = self._node.name
            content["identity"] = self._node.identity
            content["requires"] = self._node.task.requires
            content["parameters"] = self._node.task._get_parameters()

        if self._influence is not None:
            content["influence"] = self._influence
        elif self._node:
            content["influence"] = influence.HashInfluenceRegistry.get().get_strings(self._node.task)
        else:
            content["influence"] = []
        content["created"] = self._created
        content["modified"] = datetime.now()
        content["expires"] = self._expires.value
        self._created = content.get("created", datetime.now())
        self._modified = content.get("modified", datetime.now())

        ArtifactAttributeSetRegistry.format_all(self, content)

        manifest = fs.path.join(self.path, ".manifest.json")
        with open(manifest, "wb") as f:
            f.write(json.dumps(content, indent=2, default=json_serializer).encode())

    def _read_manifest(self):
        try:
            manifest_path = fs.path.join(self.path, ".manifest.json")
            with open(manifest_path) as manifest_file:
                content = json.load(manifest_file, object_hook=json_deserializer)
            self._valid = True
        except KeyboardInterrupt as e:
            raise e
        except Exception:
            self._valid = False
            return
        self._size = content["size"]
        self._unpacked = content["unpacked"]
        self._uploadable = content.get("uploadable", True)
        self._created = content.get("created", datetime.now())
        self._modified = content.get("modified", datetime.now())
        self._influence = content.get("influence", [])
        self._expires = ArtifactEvictionStrategyRegister.get().find(
            content.get("expires", "immediately"))
        ArtifactAttributeSetRegistry.parse_all(self, content)
        return content

    def _get_size(self):
        counted_inodes = {}
        size = 0
        for path, dirs, files in os.walk(self.path):
            for file in files:
                fp = os.path.join(path, file)
                try:
                    stat = os.lstat(fp)
                except OSError:
                    continue

                try:
                    counted_inodes[stat.st_ino]
                except KeyError:
                    counted_inodes[stat.st_ino] = True
                else:
                    continue

                size += stat.st_size

            for dir in dirs:
                fp = os.path.join(path, dir)
                try:
                    stat = os.lstat(fp)
                except OSError:
                    continue

                size += stat.st_size

        self._size = size
        return size

    def apply(self):
        pass

    def unapply(self):
        pass

    def is_main(self):
        return self._main

    def is_session(self):
        return self._session

    def is_valid(self):
        return self._valid

    def reload(self):
        self._unpacked = False
        self._uploadable = True
        self._created = datetime.now()
        self._modified = datetime.now()
        self._expires = self._task.expires if not self._session else expires.Immediately()
        self._size = 0
        self._influence = None
        self._valid = False
        self._temporary = False
        self._read_manifest()
        self._temporary = not self._valid

    @property
    def name(self):
        """ str: The name of the artifact. Default: 'main'. """
        return self._name

    @property
    def path(self):
        """ str: The current location of the artifact in the local cache. """
        return self._temp if self._temporary else self._path

    @property
    def final_path(self):
        """ str: The final location of the artifact in the local cache. """
        return self._path

    @property
    def stable_path(self):
        """ Deprecated. Use final_path. """
        return self._path

    @property
    def temporary_path(self):
        return self._temp

    @property
    def tools(self):
        return self._tools

    def collect(self, files, dest=None, flatten=False, symlinks=False, cwd=None):
        """ Collect files to be included in the artifact.

        Args:
            files (str): A filename pattern matching the files to be included
                in the artifact. The pattern may contain simple shell-style
                wildcards such as '*' and '?'. Note: files starting with a
                dot are not matched by these wildcards.
            dest (str, optional): Destination path within the artifact. If
                the string ends with a path separator a new directory will
                be created and all matched source files will be copied into
                the new directory. A destination without trailing path
                separator can be used to rename single files, one at a time.
            flatten (boolean, optional): If True, the directory tree structure
                of matched source files will flattened, i.e. all files will
                be copied into the root of the destination path. The default
                is False, which retains the directory tree structure
                relative to the current working directory.
            symlinks (boolean, optional): If True, symlinks are copied.
                The default is False, i.e. the symlink target is copied.
            cwd (str, optional): Change current working directory before
                starting collection.

        """
        if cwd:
            with self.tools.cwd(cwd):
                return self.collect(files, dest, flatten, symlinks)

        raise_task_error_if(
            not self.is_temporary(),
            self._node,
            "Can't collect files into an already published task artifact")

        files = self.tools.expand_path(files)
        files = self.tools.glob(files)
        dest = self.tools.expand_relpath(dest, self.tools.getcwd()) if dest is not None else ""

        # Special case for renaming files
        safe_dest = dest or fs.sep
        if len(files) == 1 and safe_dest[-1] != fs.sep and safe_dest[-1] != '.':
            src = files[0]
            self.files.append(self.tools.expand_relpath(src), dest)
            self.tools.copy(src, fs.path.join(self._temp, dest), symlinks=symlinks)
            log.verbose("Collected {0} -> {2}/{1}", src, dest, self._temp)
            return [dest]

        # Expand directories into full file list if flatting a tree
        # Determine relative artifact destination paths
        if flatten:
            files = [q
                     for f in files
                     for q in ([p for p in fs.scandir(fs.path.join(self.tools.getcwd(), f))]
                               if fs.path.isdir(fs.path.join(self.tools.getcwd(), f)) else [f])]
            reldestfiles = [fs.path.join(dest, fs.path.basename(f)) for f in files]
        else:
            reldestfiles = [fs.path.join(dest, self.tools.expand_relpath(f, self.tools.getcwd()))
                            for f in files]

        # General case
        for srcpath, reldstpath in zip(files, reldestfiles):
            relsrcpath = self.tools.expand_relpath(srcpath, self.tools.getcwd())
            dstpath = fs.path.join(self._temp, reldstpath)

            if symlinks or fs.path.exists(srcpath):
                self.files.append(self.tools.expand_relpath(srcpath), reldstpath)
                self.tools.copy(srcpath, dstpath, symlinks=symlinks)
                log.verbose("Collected {0} -> {1}", relsrcpath, reldstpath)

        return reldestfiles

    def copy(self, files, dest, flatten=False, symlinks=False, cwd=None):
        """ Copy files from the artifact.

        Args:
            files (str): A filename pattern matching the files to be copied
                from the artifact. The filepath is relative to the artifact
                root and may contain simple shell-style wildcards such as
                '*' and '?'. Note: files starting with a dot are not matched
                by these wildcards.
            dest (str, optional): Destination path, relative to the current
                working directory. If the string ends with a path separator
                a new directory will be created and all matched source files
                will be copied into the new directory. A destination without
                trailing path separator can be used to rename single files,
                one at a time.
            flatten (boolean, optional): If True, the directory tree structure
                of matched source files will flattened, i.e. all files will
                be copied into the root of the destination path. The default
                is False, which retains the directory tree structure.
            symlinks (boolean, optional): If True, symlinks are copied.
                The default is False, i.e. the symlink target is copied.
            cwd (str, optional): Change destination working directory before
                starting copy.

        """
        if cwd:
            with self.tools.cwd(cwd):
                return self.copy(files, dest, flatten, symlinks)

        raise_task_error_if(
            self.is_temporary(),
            self._node,
            "Can't copy files from an unpublished task artifact")

        files = fs.path.join(self._path, files)
        files = self.tools.expand_path(files)
        files = self.tools.glob(files)
        dest = self.tools.expand_relpath(dest, self.tools.getcwd()) if dest is not None else ""

        # Special case for renaming files
        safe_dest = dest or fs.sep
        if len(files) == 1 and safe_dest[-1] != fs.sep and safe_dest[-1] != '.':
            src = files[0]
            self.tools.copy(src, dest, symlinks=symlinks)
            log.verbose("Copied {0} -> {1}",
                        self.tools.expand_relpath(src, self._path), dest)
            return

        # Expand directories into full file list if flatting a tree
        # Determine relative artifact destination paths
        if flatten:
            files = [q
                     for f in files
                     for q in ([p for p in fs.scandir(f)] if fs.path.isdir(f) else [f])]
            reldestfiles = [fs.path.join(dest, fs.path.basename(f)) for f in files]
        else:
            reldestfiles = [fs.path.join(dest, self.tools.expand_relpath(f, self._path))
                            for f in files]

        # General case
        for srcpath, reldstpath in zip(files, reldestfiles):
            relsrcpath = self.tools.expand_relpath(srcpath, self._path)
            dstpath = fs.path.join(self.tools.getcwd(), reldstpath)

            if symlinks or fs.path.exists(srcpath):
                self.tools.copy(srcpath, dstpath, symlinks=symlinks)
                log.verbose("Copied {0} -> {1}", relsrcpath, reldstpath)

    def _set_uploadable(self, uploadable):
        self._uploadable = uploadable

    def _set_unpacked(self, unpacked=True):
        self._unpacked = unpacked

    def get_archive(self):
        return self._archive if fs.path.exists(self._archive) else None

    def get_archive_path(self):
        return self._archive

    def get_lock_path(self):
        return self._lock_path

    def get_temporary_path(self):
        return self._temp

    def get_size(self):
        return self._size

    def get_cache(self):
        return self._cache

    def get_task(self):
        return self._node.task

    def get_node(self):
        return self._node

    def is_temporary(self):
        return self._temporary

    def is_unpackable(self):
        if not self._node:
            return True
        return self._node.is_unpackable()

    def is_unpacked(self):
        return self._unpacked

    def is_uploadable(self):
        return self._uploadable

    def is_cacheable(self):
        if not self._node:
            return True
        if self.is_session():
            return True
        return self.task.is_cacheable()

    @property
    def identity(self):
        return self._identity

    @property
    def task(self):
        if not self._node:
            Task = namedtuple('Point', ['name'])
            return Task(name=self.name)
        return self._node.task


class Context(object):
    """
    Execution context and dependency wrapper.

    A ``Context`` gathers dependencies and initializes the environment
    for an executing task. It is passed as an argument to the
    Task's :func:`~jolt.Task.run` method.

    A task implementor can use the context as a dictionary of dependencies where
    the key is the name of a dependency and the value is the dependency's
    :class:`~jolt.Artifact`.

    """

    def __init__(self, cache, node):
        self._cache = cache
        self._node = node
        self._artifacts = OrderedDict()
        self._artifacts_index = OrderedDict()

    def __enter__(self):
        try:
            for dep in reversed(self._node.children):
                for artifact in dep.artifacts:
                    # Create clone with tools from this task
                    artifact = self._cache.get_artifact(
                        dep,
                        name=artifact.name,
                        session=artifact.is_session(),
                        tools=self._node.tools,
                    )

                    # Don't include session artifacts that don't exist,
                    # i.e. were no build has taken place due to presence
                    # of the persistent artifacts.
                    if artifact.is_session() and not self._cache.is_available_locally(artifact):
                        continue

                    self._cache.unpack(artifact)

                    if artifact.name == "main":
                        self._artifacts_index[dep.qualified_name] = artifact
                        self._artifacts_index[dep.short_qualified_name] = artifact
                    self._artifacts[artifact.name + "@" + dep.qualified_name] = artifact
                    self._artifacts_index[artifact.name + "@" + dep.qualified_name] = artifact
                    self._artifacts_index[artifact.name + "@" + dep.short_qualified_name] = artifact
                    artifact.apply()
                    ArtifactAttributeSetRegistry.apply_all(self._node.task, artifact)
            ArtifactAttributeSetRegistry.apply_all_deps(self._node.task, self)
        except Exception as e:
            # Rollback all attributes/resources except the last failing one
            ArtifactAttributeSetRegistry.unapply_all_deps(self._node.task, self)
            for name, artifact in reversed(list(self._artifacts.items())[:-1]):
                with utils.ignore_exception():
                    ArtifactAttributeSetRegistry.unapply_all(self._node.task, artifact)
                    artifact.unapply()
            raise e
        return self

    def __exit__(self, type, value, tb):
        ArtifactAttributeSetRegistry.unapply_all_deps(self._node.task, self)
        for name, artifact in reversed(self._artifacts.items()):
            ArtifactAttributeSetRegistry.unapply_all(self._node.task, artifact)
            artifact.unapply()

    def __getitem__(self, key):
        """ Get artifact for a task listed as a requirement.

        Args:
            key (str): Name of the task listed as a requirement.

        Returns:
            The :class:`~jolt.Artifact` associated with the task.

        Example:
            .. code-block:: python

                requires = "dependency"

                def run(self, deps, tools):
                    dependency_artifact = deps["dependency"]

        """

        key = self._node.tools.expand(key)

        alias, artifact, task, params = utils.parse_aliased_task_name(key)
        raise_task_error_if(alias, self._node, "Cannot define alias when indexing dependencies")
        task_name = utils.format_task_name(task, params)
        task_artifact_name = utils.format_task_name(task, params, artifact)

        if task_name not in self._artifacts_index and \
           task_artifact_name not in self._artifacts_index and not params:
            key = self._node.resolve_requirement_alias(task_name)
            if key:
                _, _, task, params = utils.parse_aliased_task_name(key)
                task_name = utils.format_task_name(task, params)
                task_artifact_name = utils.format_task_name(task, params, artifact)

        # Parameters may be overspecified, resolve task
        if task_artifact_name not in self._artifacts_index:
            from jolt.tasks import TaskRegistry
            task_obj = TaskRegistry.get().get_task(task_name)
            task_name = task_obj.short_qualified_name if task_obj is not None else task
            task_artifact_name = task_name if not artifact else f"{artifact}@{task_name}"

        if task_artifact_name not in self._artifacts_index:
            raise KeyError("No such artifact dependency '{0}' ({1})".format(
                task_artifact_name, self._node.short_qualified_name))
        return self._artifacts_index[task_artifact_name]

    def items(self):
        """ List all requirements and their artifacts.

        Returns:
            Requirement dictionary items. Each item is a tuple with the
            requirement name and the artifact.
        """
        return reversed(self._artifacts.items())


class PidProvider(object):
    def __init__(self):
        self._uuid = uuid.uuid4()

    def __call__(self):
        return str(self._uuid)


@utils.Singleton
class ArtifactCache(StorageProvider):
    """
    Manages the local artifact cache and exchanges artifacts with
    external caches using configured storage provider plugins.

    Artifacts are directories containing files published by tasks.
    They are normally stored uncompressed in the filesystem for
    fast access during builds. They are exchanged with external
    caches as gzip compressed tarballs.

    Unused artifacts can be evicted when new artifacts are committed
    to the cache if the configured cache size is exceeded. Selection
    follows LRU order, but deviations are possible through artifact
    eviction policies. For example, an important large artifact could
    declare that it shouldn't be evicted unless unused for two weeks.
    It would then not be considered for eviction until later.

    Artifacts in the cache can be accessed by multiple processes in
    parallel. Critical sections are enforced using a combination of
    file locks and database record keeping.

    Each locally available artifact is recorded in an Sqlite database
    (table: artifacts).

    If an artifact is required during a build, an artifact reference
    is recorded in the database table ``artifact_refs``. This prevents
    the artifact from being evicted while the process is alive. The
    reference is garbage collected by other processes if the owner
    process should terminate abnormally.

    If an artifact is not present in the cache and either has to be
    built or downloaded, the artifact will be interprocess locked,
    i.e. it cannot be built or downloaded by any other process.
    Artifacts are also locked when the Task.unpack() method is run.

    Each process that wishes to lock an artifact records their
    interest in the database table ``artifact_lockrefs``. Actual
    interprocess concurrency is prevented by using a file lock
    in <cache_directory>/locks/<artifact_identity>.lock.
    This lock file is deleted by the last process holding lock reference
    in the database, or when garbage collected.

    Each process also owns a pid lock file in
    <cache_directory>/locks/<pid>.lock. The purpose of this file is
    to allow other processes to detect termination of the owner
    process. If a read lock cannot be obtained on this file then the
    process is still alive. Otherwise, any artifact references created
    by the owner process may be garbage collected.

    A global cache lock is always acquired when creating or deleting
    artifact or process lock files. It's also used when performing
    database transactions for simplicity.
    """

    storage_provider_factories = []

    def __init__(self, options=None, pidprovider=None):
        self._options = options or JoltOptions()
        self._storage_providers = [
            factory.create(self)
            for factory in ArtifactCache.storage_provider_factories]

        # If no storage providers supports the availability method,
        # we will not only use the local presence cache.
        self._local_presence_cache = set()
        self._remote_presence_cache = set()
        self._presence_cache_only = self.has_availability()

        # Read configuration
        self._max_size = config.getsize(
            "jolt", "cachesize", os.environ.get("JOLT_CACHESIZE", 1 * 1024 ** 3))

        # Create cache directory
        self._fs_create_cachedir()

        # Create global cache lock file
        self._cache_locked = False
        self._lock_file = fasteners.InterProcessLock(self._fs_get_lock_file())
        self._thread_lock = RLock()

        # Create process lock file
        with self._cache_lock():
            self._pid = pidprovider() if pidprovider else PidProvider()()
            self._pid_file = fasteners.InterProcessLock(self._fs_get_pid_file(self._pid))
            self._pid_file.acquire()

        # Setup database and garbage collect stale refs
        self._db_path = self._fs_get_db_path()
        with self._cache_lock(), self._db() as db:
            self._db_create_tables(db)
            self._db_invalidate_locks(db)
            self._db_invalidate_references(db)
            self._fs_invalidate_pids(db)
            size = self._db_select_sum_artifact_size(db)
            count = self._db_select_artifact_count(db)
            in_use = self._db_select_artifact_count_in_use(db)
            cur_size = utils.as_human_size(size)
            max_size = utils.as_human_size(self._max_size)
            log.verbose("Cache size is {} (max {}, {} artifacts, {} in use)",
                        cur_size, max_size, count, in_use)
        atexit.register(self.close)

    ############################################################################
    # Internal API
    ############################################################################

    def _assert_cache_locked(self):
        assert self._cache_locked, "illegal function call, cache lock is not held"

    @contextlib.contextmanager
    def _db(self):
        db = sqlite3.connect(self._db_path, detect_types=sqlite3.PARSE_DECLTYPES)
        try:
            db.execute("PRAGMA journal_mode=OFF")
            yield db
        finally:
            db.close()

    def _db_create_tables(self, db):
        cur = db.cursor()

        # All artifacts currently residing in the cache
        cur.execute("CREATE TABLE IF NOT EXISTS artifacts "
                    "(identity text PRIMARY KEY, name text, size integer, last_used timestamp)")

        # All process references to artifacts in the cache. No eviction allowed while rows exist here.
        cur.execute("CREATE TABLE IF NOT EXISTS artifact_refs (identity text, pid text)")

        # All process references to artifict locks.
        # A lock may exist before the artifact, for example during building or downloading.
        # A lock file may be safely deleted if the global cache lock is held and there are
        # no rows present.
        cur.execute("CREATE TABLE IF NOT EXISTS artifact_lockrefs (identity text, pid text)")
        db.commit()

    def _db_insert_artifact(self, db, identity, task_name, size):
        cur = db.cursor()
        cur.execute("INSERT INTO artifacts VALUES (?,?,?,?)", (identity, task_name, size, datetime.now()))
        db.commit()

    def _db_update_artifact_size(self, db, identity, size):
        cur = db.cursor()
        cur.execute("UPDATE artifacts SET size = ? WHERE identity = ?", (size, identity))
        db.commit()

    def _db_delete_artifact(self, db, identity, and_refs=True):
        cur = db.cursor()
        if and_refs:
            cur.execute("DELETE FROM artifact_refs WHERE identity = ?", (identity,))
        cur.execute("DELETE FROM artifacts WHERE identity = ?", (identity,))
        db.commit()

    def _db_insert_reference(self, db, identity):
        cur = db.cursor()
        cur.execute("INSERT INTO artifact_refs VALUES (?,?)", (identity, self._pid))
        cur.execute("UPDATE artifacts SET last_used = ? WHERE identity = ?", (datetime.now(), identity))
        db.commit()

    def _db_delete_reference(self, db, identity):
        cur = db.cursor()
        cur.execute("DELETE FROM artifact_refs WHERE identity = ? AND pid = ?", (identity, self._pid))
        db.commit()

    def _db_select_reference(self, db, identity):
        cur = db.cursor()
        return list(cur.execute("SELECT * FROM artifact_refs WHERE identity = ? AND pid = ?", (identity, self._pid)))

    def _db_insert_lock(self, db, identity):
        cur = db.cursor()
        cur.execute("INSERT INTO artifact_lockrefs VALUES (?,?)", (identity, self._pid))
        db.commit()

    def _db_delete_lock(self, db, identity):
        cur = db.cursor()
        cur.execute("DELETE FROM artifact_lockrefs WHERE identity = ? AND pid = ?", (identity, self._pid))
        db.commit()

    def _db_delete_locks_by_pid(self, db, pid):
        cur = db.cursor()
        cur.execute("DELETE FROM artifact_lockrefs WHERE pid = ?", (pid,))
        db.commit()

    def _db_select_locks(self, db):
        cur = db.cursor()
        return [n[0] for n in cur.execute("SELECT DISTINCT identity FROM artifact_lockrefs")]

    def _db_select_lock_count(self, db, identity):
        cur = db.cursor()
        record = cur.execute("SELECT COUNT(*) FROM artifact_lockrefs WHERE identity = ?",
                             (identity,)).fetchone()
        return record[0]

    def _db_delete_references_by_pid(self, db, pid):
        cur = db.cursor()
        cur.execute("DELETE FROM artifact_refs WHERE pid = ?", (pid,))
        db.commit()

    def _db_select_artifact(self, db, identity):
        cur = db.cursor()
        return list(cur.execute("SELECT * FROM artifacts WHERE identity = ?", (identity,)))

    def _db_select_artifacts(self, db):
        cur = db.cursor()
        return list(cur.execute("SELECT * FROM artifacts"))

    def _db_select_lock_pids(self, db):
        cur = db.cursor()
        return [n[0] for n in cur.execute("SELECT DISTINCT pid FROM artifact_lockrefs")]

    def _db_select_artifact_lock_pids(self, db, identity):
        cur = db.cursor()
        return [n[0] for n in cur.execute("SELECT DISTINCT pid FROM artifact_lockrefs WHERE identity = ?", (identity,))]

    def _db_select_reference_pids(self, db):
        cur = db.cursor()
        return [n[0] for n in cur.execute("SELECT DISTINCT pid FROM artifact_refs")]

    def _db_select_artifact_reference_pids(self, db, identity):
        cur = db.cursor()
        return [n[0] for n in cur.execute("SELECT DISTINCT pid FROM artifact_refs WHERE identity = ?", (identity,))]

    def _db_select_artifact_not_in_use(self, db, identity):
        cur = db.cursor()
        return list(
            cur.execute("SELECT * FROM artifacts WHERE identity = ? AND identity NOT IN "
                        "(SELECT identity FROM artifact_refs) "
                        "ORDER BY last_used", (identity,)))

    def _db_select_artifacts_not_in_use(self, db):
        cur = db.cursor()
        return list(
            cur.execute("SELECT * FROM artifacts WHERE identity NOT IN "
                        "(SELECT identity FROM artifact_refs) "
                        "ORDER BY last_used"))

    def _db_select_sum_artifact_size(self, db):
        cur = db.cursor()
        return list(cur.execute("SELECT SUM(size) FROM artifacts"))[0][0] or 0

    def _db_select_artifact_count(self, db):
        cur = db.cursor()
        return list(cur.execute("SELECT COUNT(identity) FROM artifacts"))[0][0] or 0

    def _db_select_artifact_count_in_use(self, db):
        cur = db.cursor()
        return list(cur.execute("SELECT COUNT(DISTINCT artifact_refs.identity) FROM artifact_refs"))[0][0] or 0

    def _db_invalidate_locks(self, db, try_all=False):
        """ Removes any stale artifact lock references and lock files """
        self._assert_cache_locked()
        locks = self._db_select_locks(db)
        for pid in self._db_select_lock_pids(db):
            if not try_all and pid == self._pid:
                continue
            try:
                # Throws exception if lock is held
                with self._pid_lock(pid):
                    self._db_delete_locks_by_pid(db, pid)
            except KeyboardInterrupt as e:
                raise e
            except Exception:
                pass

        for lock in locks:
            if self._db_select_lock_count(db, lock) == 0:
                fs.unlink(self._fs_get_artifact_lockpath(lock), ignore_errors=True)

    def _db_invalidate_references(self, db, try_all=False):
        """ Removes any stale artifact references """
        self._assert_cache_locked()
        for pid in self._db_select_reference_pids(db):
            if not try_all and pid == self._pid:
                continue
            try:
                # Throws exception if lock is held
                with self._pid_lock(pid):
                    self._db_delete_references_by_pid(db, pid)
            except KeyboardInterrupt as e:
                raise e
            except Exception:
                pass

    def _fs_invalidate_pids(self, db, try_all=False):
        """ Removes any stale pid files """
        self._assert_cache_locked()
        for pid in fs.scandir(fs.path.join(self.root, "pids")):
            if not try_all and fs.path.basename(pid) == self._pid:
                continue
            try:
                # Throws exception if lock is held
                with self._pid_lock(pid):
                    pass
                fs.unlink(pid, ignore_errors=True)
            except KeyboardInterrupt as e:
                raise e
            except Exception:
                pass

    def _fs_create_cachedir(self):
        self.root = config.get_cachedir()
        log.verbose("Jolt cache path: {}", self.root)
        try:
            fs.makedirs(self.root)
        except KeyboardInterrupt as e:
            raise e
        except Exception:
            raise_error("Failed to create cache directory '{0}'", self.root)

    def _fs_get_artifact(self, node, name, tools=None, session=False):
        return Artifact(self, node, name=name, tools=tools, session=session)

    def _fs_commit_artifact(self, artifact, uploadable):
        artifact._set_uploadable(uploadable)
        if not artifact.is_unpackable():
            artifact._set_unpacked()
        artifact._write_manifest()
        if artifact.is_temporary():
            fs.rmtree(artifact.final_path, ignore_errors=True)
            fs.rename(artifact.path, artifact.final_path)

    @contextlib.contextmanager
    def _fs_compress_artifact(self, artifact):
        task = artifact.task
        archive = artifact.get_archive_path()

        raise_task_error_if(
            artifact.is_temporary(), task,
            "Can't compress an unpublished task artifact")

        try:
            artifact.tools.archive(artifact.path, archive)
        except KeyboardInterrupt as e:
            raise e
        except Exception:
            raise_task_error(task, "Failed to compress task artifact")
        try:
            yield
        finally:
            fs.unlink(archive, ignore_errors=True)

    def _fs_decompress_artifact(self, artifact):
        task = artifact.task
        archive = artifact.get_archive_path()
        try:
            task.tools.extract(archive, artifact.temporary_path, ignore_owner=True)
        except KeyboardInterrupt as e:
            raise e
        except Exception:
            raise_task_error(task, "Failed to extract task artifact archive")
        finally:
            fs.unlink(archive, ignore_errors=True)
        artifact._read_manifest()

    def _fs_delete_artifact(self, identity, task_name, onerror=None):
        fs.rmtree(self._fs_get_artifact_path(identity, task_name), ignore_errors=True, onerror=onerror)
        fs.rmtree(self._fs_get_artifact_tmppath(identity, task_name), ignore_errors=True, onerror=onerror)
        fs.unlink(fs.path.join(self.root, task_name), ignore_errors=True)

    def _fs_get_artifact_archivepath(self, identity, task_name):
        return fs.get_archive(fs.path.join(self.root, task_name, identity))

    def _fs_get_artifact_lockpath(self, identity):
        return fs.path.join(self.root, "locks", identity + ".lock")

    def _fs_get_artifact_tmppath(self, identity, task_name):
        return fs.path.join(self.root, task_name, "." + identity)

    def _fs_get_artifact_path(self, identity, task_name):
        return fs.path.join(self.root, task_name, identity)

    def _fs_get_artifact_manifest_path(self, identity, task_name):
        return fs.path.join(self._fs_get_artifact_path(identity, task_name), ".manifest.json")

    def _fs_get_artifact_manifest(self, identity, task_name):
        path = self._fs_get_artifact_manifest_path(identity, task_name)
        with open(path) as manifest_file:
            return json.load(manifest_file, object_hook=json_deserializer)

    def _fs_get_db_path(self):
        return fs.path.join(self.root, "cache.db")

    def _fs_get_lock_file(self):
        return fs.path.join(self.root, "lock")

    def _fs_get_pid_file(self, pid):
        return fs.path.join(self.root, "pids", pid)

    def _fs_is_artifact_expired(self, identity, task_name, last_used):
        try:
            manifest = self._fs_get_artifact_manifest(identity, task_name)
            manifest["used"] = last_used
            strategy = ArtifactEvictionStrategyRegister.get().find(
                manifest.get("expires", "immediately"))
            return strategy.is_evictable(manifest)
        except KeyboardInterrupt as e:
            raise e
        except Exception:
            return True

    def close(self):
        with self._cache_lock(), self._db() as db:
            self._db_invalidate_locks(db, try_all=True)
            self._db_invalidate_references(db, try_all=True)
            self._fs_invalidate_pids(db, try_all=True)

    @contextlib.contextmanager
    def _cache_lock(self):
        with self._thread_lock:
            self._lock_file.acquire()
            self._cache_locked = True
            try:
                yield
            finally:
                self._cache_locked = False
                self._lock_file.release()

    @contextlib.contextmanager
    def _pid_lock(self, pid, wait=False, timeout=None):
        """
        Process specific lock.

        Indicator of process liveness. Throws exception if process is alive.
        """
        if not wait:
            self._assert_cache_locked()
        with self._thread_lock:
            lock_file = self._fs_get_pid_file(pid)
            lock = fasteners.InterProcessLock(lock_file)
            if not lock.acquire(blocking=wait, timeout=timeout):
                raise RuntimeError()
            try:
                yield
            finally:
                lock.release()

    def _discard(self, db, artifacts, if_expired, onerror=None):
        """ Discard list of artifacts. Cache lock must be held. """
        self._assert_cache_locked()
        evicted = 0
        for identity, task_name, _, used in artifacts:
            if not if_expired or self._fs_is_artifact_expired(identity, task_name, used):
                self._db_delete_artifact(db, identity)
                self._fs_delete_artifact(identity, task_name, onerror=onerror)
                evicted += 1
                log.debug("Evicted {}: {}", identity, task_name)
        return evicted == len(artifacts)

    ############################################################################
    # Public API
    ############################################################################

    def is_available_locally(self, artifact):
        """
        Check presence of task artifact in cache.

        If the artifact is present, an artifact reference is automatically
        recorded for the running process to prevent eviction by other
        processes.
        """
        if not artifact.is_cacheable():
            return False

        # Cache availability in node
        try:
            assert artifact.identity in self._local_presence_cache
        except AssertionError:
            pass

        with self._cache_lock(), self._db() as db:
            if self._db_select_artifact(db, artifact.identity) or self._db_select_reference(db, artifact.identity):
                artifact.reload()
                if artifact.is_temporary():
                    self._db_delete_artifact(db, artifact.identity)
                    return False
                self._db_insert_reference(db, artifact.identity)
                self._local_presence_cache.add(artifact.identity)
                return True
        return False

    def is_available_remotely(self, artifact, cache=True):
        """
        Check presence of task artifact in external remote caches.
        """
        if cache:
            if artifact.identity in self._remote_presence_cache:
                return True
            if self._presence_cache_only:
                return False
        for provider in self._storage_providers:
            present, _ = provider.availability([artifact])
            if present:
                self._remote_presence_cache.add(artifact.identity)
                return True
        return False

    def is_available(self, artifact):
        """ Check presence of task artifact in any cache, local or remote """
        return self.is_available_locally(artifact) or self.is_available_remotely(artifact)

    def has_availability(self):
        # Returns true if all storage providers implement the availability method
        return all([provider.availability.__func__ != StorageProvider.availability for provider in self._storage_providers])

    def availability(self, artifacts, remote=True):
        """ Check presence of task artifacts in any cache, local or remote """
        present = set()
        missing = set()

        # Make sure artifacts is a list
        artifacts = utils.as_list(artifacts)

        # Check presence of all artifacts in the local cache
        for artifact in artifacts:
            if self.is_available_locally(artifact):
                present.add(artifact)
            else:
                missing.add(artifact)

        if not remote:
            return list(present), list(missing)

        # Check presence of all artifacts in the remote caches
        missing_remotely = artifacts

        for provider in self._storage_providers:
            present_in_provider, missing_in_provider = provider.availability(missing_remotely)
            for artifact in present_in_provider:
                self._remote_presence_cache.add(artifact.identity)
            present.update(present_in_provider)
            missing_remotely = missing_in_provider
            if not missing_in_provider:
                break

        missing.update(missing_remotely)
        missing = missing - present

        return list(present), list(missing)

    def download_enabled(self):
        return self._options.download and \
            any([provider.download_enabled() for provider in self._storage_providers])

    def upload_enabled(self):
        return self._options.upload and \
            any([provider.upload_enabled() for provider in self._storage_providers])

    def download_all(self, node, force=False):
        pass

    def download(self, artifact, force=False):
        """
        Downloads an artifact from a remote cache to the local cache.

        The artifact is interprocess locked during the operation.
        """
        if not force and not self.download_enabled():
            return False
        if not artifact.is_cacheable():
            return False
        with self.lock_artifact(artifact) as artifact:
            if self.is_available_locally(artifact):
                artifact.task.info("Download skipped, already in local cache ({name})")
                return True
            for provider in self._storage_providers:
                if provider.download(artifact, force):
                    self._fs_decompress_artifact(artifact)
                    self.commit(artifact)
                    return True
        return len(self._storage_providers) == 0

    def upload(self, artifact, force=False, locked=True):
        """
        Uploads an artifact from the local cache to all configured remote caches.

        The artifact is interprocess locked during the operation.
        """
        if not force and not self.upload_enabled():
            return False
        if not artifact.is_cacheable():
            return True
        raise_task_error_if(
            not self.is_available_locally(artifact), artifact.task,
            "Can't upload task artifact, no artifact present in the local cache")
        with self.lock_artifact(artifact) if locked else artifact as artifact:
            raise_task_error_if(
                not artifact.is_uploadable(), artifact.task,
                "Artifact was modified locally by another process and can no longer be uploaded, try again")
            if self._storage_providers:
                with self._fs_compress_artifact(artifact):
                    return all([provider.upload(artifact, force) for provider in self._storage_providers])
        return len(self._storage_providers) == 0

    def location(self, artifact):
        for provider in self._storage_providers:
            url = provider.location(artifact)
            if url:
                return url
        return ''

    def unpack(self, artifact):
        """
        Unpacks/relocates the task artifact to the local cache.

        A temporary backup of the artifact is first created in the
        filesystem. The task's unpack() method is then executed with
        original artifact path as argument (to allow relocation to
        the final artifact path). If unpack() succeeds, the temporary
        backup is discarded. Otherwise the backup is restored.

        The artifact is interprocess locked during the operation.
        """
        if not artifact.is_unpackable():
            return True
        with self._thread_lock, self.lock_artifact(artifact) as artifact:
            if not self.is_available_locally(artifact):
                raise_task_error(
                    artifact.task,
                    "Locked artifact is missing in cache (forcibly removed?)")
            if artifact.is_unpacked():
                return True

            # Keep a temporary copy of the artifact if the task
            # unpack() method fails. The copy is removed in
            # get_locked_artifact() if left unused.
            fs.copy(artifact.path, artifact.temporary_path, symlinks=True)

            task = artifact.task
            with tools.Tools(task) as t:
                try:
                    # Note: unpack() will run on the original
                    # artifact, not in the temporary copy.
                    if task.unpack.__func__ is not tasks.Task.unpack:
                        task.info("Unpack started {}", artifact.get_node().log_name)
                    artifact._set_unpacked()
                    task.unpack(artifact, t)
                    self.commit(artifact, uploadable=False)

                except NotImplementedError:
                    self.commit(artifact)

                except Exception as e:
                    # Restore the temporary copy
                    fs.rmtree(artifact.path, ignore_errors=True)
                    fs.rename(artifact.temporary_path, artifact.path)
                    artifact.task.error("Unpack failed {}", artifact.get_node().log_name)
                    raise e
        return True

    def commit(self, artifact, uploadable=True):
        """
        Commits a task artifact to the cache.

        Committing includes renaming the artifact in the filesystem,
        adding an artifact database record as well as a process reference
        record.

        Once the artifact is committed, eviction of other artifacts will
        take place if the resulting cache size exceeds the configured
        limit.
        """
        if not artifact.is_cacheable():
            return

        with self._cache_lock(), self._db() as db:
            self._fs_commit_artifact(artifact, uploadable)
            with utils.ignore_exception():  # Possibly already exists in DB, e.g. unpacked
                self._db_insert_artifact(db, artifact.identity, artifact.task.canonical_name, artifact.get_size())
            self._db_update_artifact_size(db, artifact.identity, artifact.get_size())
            self._db_insert_reference(db, artifact.identity)
            artifact.reload()

            evict_size = self._db_select_sum_artifact_size(db) - self._max_size
            if evict_size < 0:
                return

            unused = self._db_select_artifacts_not_in_use(db)
            while evict_size > 0 and unused:
                candidate, unused = unused[0], unused[1:]
                if self._discard(db, [candidate], True):
                    evict_size -= candidate[2]

    def discard(self, artifact, if_expired=False, onerror=None):
        with self._cache_lock(), self._db() as db:
            self._db_invalidate_locks(db)
            self._db_invalidate_references(db)
            self._fs_invalidate_pids(db)
            discarded = self._discard(
                db,
                self._db_select_artifact_not_in_use(db, artifact.identity),
                if_expired,
                onerror=onerror)
            if discarded:
                self._local_presence_cache.discard(artifact.identity)
            return discarded

    def _discard_wait(self, artifact):
        """
        Discards an artifact without expiration consideration.

        The artifact must be locked prior to calling this function.

        If the artifact is in use, the function waits for all references to be dropped.
        The artifact is also unregistered from the database to prevent new references,
        but it remains available in the filesystem.
        """
        with self._cache_lock(), self._db() as db:
            self._db_invalidate_locks(db)
            self._db_invalidate_references(db)
            self._fs_invalidate_pids(db)
            artifacts = self._db_select_artifact(db, artifact.identity)
            self._db_delete_artifact(db, artifact.identity, and_refs=False)
            refpids = self._db_select_artifact_reference_pids(db, artifact.identity)
            lockpids = self._db_select_artifact_lock_pids(db, artifact.identity)

        if len(refpids) > 1:
            artifact.task.info("Artifact is temporarily in use, forced discard on hold ({name})")
            for pid in refpids:
                # Loop waiting for other processes to surrender the artifact
                while True:
                    try:
                        # No need to wait for self, nor for processes waiting
                        # for the artifact lock we are already holding.
                        if pid == self._pid or pid in lockpids:
                            break
                        # Throws exception after 1s if lock is not acquired.
                        # We then we check if there are any new lock references,
                        # which allows us to a break deadlock condition.
                        with self._pid_lock(pid, wait=True, timeout=1):
                            break
                    except RuntimeError:
                        with self._cache_lock(), self._db() as db:
                            lockpids = self._db_select_artifact_lock_pids(db, artifact.identity)

        with self._cache_lock(), self._db() as db:
            assert self._discard(db, artifacts, False), "Failed to discard artifact"
            self._local_presence_cache.discard(artifact.identity)
            artifact.reload()
        return artifact

    def discard_all(self, if_expired=False, onerror=None):
        with self._cache_lock(), self._db() as db:
            self._db_invalidate_locks(db)
            self._db_invalidate_references(db)
            self._fs_invalidate_pids(db)
            return self._discard(
                db,
                self._db_select_artifacts_not_in_use(db),
                if_expired,
                onerror=onerror)

    def get_context(self, node):
        return Context(self, node)

    def get_artifact(self, node, name, tools=None, session=False):
        return self._fs_get_artifact(node, name=name, tools=tools, session=session)

    @contextlib.contextmanager
    def lock_artifact(self, artifact, discard=False):
        """
        Locks the task artifact.

        First records interest in artifact lock file and then attempts
        to acquire the lock. Deletes the file upon releasing the lock
        if there are no other references to the lock from other processes.
        """
        with self._cache_lock():
            with self._db() as db:
                self._db_insert_lock(db, artifact.identity)
                self._db_insert_reference(db, artifact.identity)
            lock_path = self._fs_get_artifact_lockpath(artifact.identity)
            lock = fasteners.InterProcessLock(lock_path)
            is_locked = lock.acquire(blocking=False)
        if not is_locked:
            artifact.task.info("Artifact is temporarily locked by another process ({name})")
            lock.acquire()

        try:
            artifact.reload()
            if discard:
                artifact = self._discard_wait(artifact)
            if artifact.is_temporary():
                fs.rmtree(artifact.temporary_path, ignore_errors=True)
                fs.makedirs(artifact.temporary_path)

            with contextlib.ExitStack() as stack:
                stack.enter_context(artifact)
                yield artifact
        finally:
            fs.rmtree(artifact.temporary_path, ignore_errors=True)
            with self._cache_lock():
                with self._db() as db:
                    self._db_delete_lock(db, artifact.identity)
                lock.release()
                with self._db() as db:
                    if self._db_select_lock_count(db, artifact.identity) == 0:
                        fs.unlink(lock_path, ignore_errors=True)

    def precheck(self, artifacts, remote=True):
        """ Precheck artifacts for availability and cache status. """
        if not self.has_availability():
            return
        present, missing = self.availability(artifacts, remote=remote)
        log.verbose("Cache: {}/{} artifacts present", len(present), len(artifacts))



import click
import os
import sys
import subprocess

from jolt import cache
from jolt import cli
from jolt import filesystem as fs
from jolt import graph
from jolt import log
from jolt import scheduler
from jolt.error import raise_task_error_if
from jolt.hooks import TaskHookRegistry
from jolt.options import JoltOptions
from jolt.plugins import ninja
from jolt.tasks import TaskRegistry, WorkspaceResource


def has_incpaths(artifact):
    return artifact.cxxinfo.incpaths.count() > 0


def stage_artifacts(artifacts, tools):
    for artifact in filter(has_incpaths, artifacts):
        tools.sandbox(artifact, incremental=True, reflect=fs.has_symlinks())


def get_task_artifacts(task):
    artifacts = []
    for dep in task.children:
        artifacts.extend(dep.artifacts)
    return task.artifacts[0], artifacts


@cli.cli.command(name="gdb", context_settings={"ignore_unknown_options": True})
@click.argument("task", type=str, shell_complete=cli._autocomplete_tasks)
@click.argument("gdb-args", type=str, nargs=-1, required=False)
@click.option("-d", "--default", type=str, multiple=True, help="Override default parameter values.")
@click.option("-mi", "--machine-interface", is_flag=True, help="Enable the machine interface for use within an IDE.")
@click.option("-nb", "--no-binary", is_flag=True,
              help="Don't load symbols from binary. Shortens startup time in IDEs where binaries are loaded through the machine interface.")
@click.pass_context
def gdb(ctx, task, default, machine_interface, no_binary, gdb_args):
    """
    Launch gdb with an executable from a task artifact.

    The executable to debug must be described in the task's artifact
    metadata using the ``artifact.strings.executable`` variable. This
    is provided by default if the task is a compilation task derived from
    :class:`ninja.CXXExecutable`.

    Before starting GDB, the command downloads all the task's dependencies
    and ensures that its workspace resources are present in the workspace.
    This allows GDB to find source files in both artifacts and SCM
    repositories.

    The command automatically sets the sysroot property in GDB
    if any of the environment variables ``SYSROOT`` or ``SDKTARGETSYSROOT``
    is set in the execution environment of the task.

    The ``-mi`` argument should typically be passed if the command is
    launched from within an IDE to allow the IDE to parse and interpret
    output from GDB.

    Additional user-defined arguments can be passed to GDB in GDB_ARGS
    immediately following the TASK name.

    """

    if machine_interface:
        log.enable_gdb()

    manifest = ctx.obj["manifest"]
    options = JoltOptions(default=default)
    acache = cache.ArtifactCache.get(options)
    TaskHookRegistry.get(options)
    executors = scheduler.ExecutorRegistry.get(options)
    registry = TaskRegistry.get()
    strategy = scheduler.DownloadStrategy(executors, acache)
    queue = scheduler.TaskQueue(strategy, acache, {})

    for params in default:
        registry.set_default_parameters(params)

    gb = graph.GraphBuilder(registry, acache, manifest, options, progress=True)
    dag = gb.build([task])

    try:
        with log.progress("Progress", dag.number_of_tasks(), " tasks", estimates=False, debug=False) as p:
            while dag.has_tasks() or not queue.empty():
                leafs = dag.select(lambda graph, task: task.is_ready())

                while leafs:
                    task = leafs.pop()
                    queue.submit(task)

                task, error = queue.wait()

                # Materialize workspace resources so that
                # source code is available to the debugger.
                if isinstance(task.task, WorkspaceResource):
                    task.task.acquire()

                p.update(1)

    except KeyboardInterrupt:
        print()
        log.warning("Interrupted by user")
        try:
            queue.abort()
            sys.exit(1)
        except KeyboardInterrupt:
            print()
            log.warning("Interrupted again, exiting")
            os._exit(1)

    except Exception as e:
        log.set_interactive(True)
        raise e

    finally:
        queue.shutdown()

    assert len(dag.goals), "Too many tasks, can only debug one executable at a time"

    for goal in dag.goals:
        main, deps = get_task_artifacts(goal)
        stage_artifacts([main] + deps, goal.tools)

        raise_task_error_if(
            main.strings.executable.get_value() is None,
            goal, "No executable found in task artifact")

        with acache.get_context(goal):
            gdb = goal.tools.getenv("GDB", "gdb")
            cmd = [gdb]
            sysroot = goal.tools.getenv("SDKTARGETSYSROOT", goal.tools.getenv("SYSROOT"))
            if sysroot:
                cmd += ["-ex", "set sysroot " + sysroot]
                cmd += ["-ex", "add-auto-load-safe-path " + sysroot]
            if fs.has_symlinks():
                cmd += ["-ex", "set substitute-path ../sandboxes/ ../sandboxes-reflected/"]
            if machine_interface:
                cmd += ["-i=mi"]
            cmd += ["-ex", "set print asm-demangle"]
            cmd += ["-ex", "set print thread-events off"]
            cmd += ["-ex", "handle SIG32 nostop noprint"]
            if not no_binary:
                cmd += [os.path.join(main.path, str(main.strings.executable))]
            cmd += gdb_args

            if isinstance(goal.task, ninja.CXXProject):
                cwd = goal.tools.builddir("ninja", incremental=True)
            else:
                cwd = goal.task.joltdir

            with goal.tools.environ() as env:
                subprocess.call(cmd, env=env, cwd=cwd)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jolt/plugins/remote_execution/scheduler.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from jolt import common_pb2 as jolt_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n-jolt/plugins/remote_execution/scheduler.proto\x1a\x11jolt/common.proto\"[\n\x0c\x42uildRequest\x12&\n\x0b\x65nvironment\x18\x01 \x01(\x0b\x32\x11.BuildEnvironment\x12\x10\n\x08priority\x18\x02 \x01(\x05\x12\x11\n\tlogstream\x18\x03 \x01(\x08\"=\n\x0b\x42uildUpdate\x12\x1c\n\x06status\x18\x01 \x01(\x0e\x32\x0c.BuildStatus\x12\x10\n\x08\x62uild_id\x18\x02 \x01(\t\"&\n\x12\x43\x61ncelBuildRequest\x12\x10\n\x08\x62uild_id\x18\x01 \x01(\t\"3\n\x13\x43\x61ncelBuildResponse\x12\x1c\n\x06status\x18\x01 \x01(\x0e\x32\x0c.BuildStatus\"0\n\x0bTaskRequest\x12\x10\n\x08\x62uild_id\x18\x01 \x01(\t\x12\x0f\n\x07task_id\x18\x02 \x01(\t\"0\n\x10WorkerAllocation\x12\n\n\x02id\x18\x01 \x01(\t\x12\x10\n\x08hostname\x18\x02 \x01(\t\"\xa3\x01\n\nTaskUpdate\x12\x1d\n\x07request\x18\x01 \x01(\x0b\x32\x0c.TaskRequest\x12\x1b\n\x06status\x18\x02 \x01(\x0e\x32\x0b.TaskStatus\x12!\n\x06worker\x18\x03 \x01(\x0b\x32\x11.WorkerAllocation\x12\x1a\n\x08loglines\x18\x04 \x03(\x0b\x32\x08.LogLine\x12\x1a\n\x06\x65rrors\x18\x05 \x03(\x0b\x32\n.TaskError2\xa2\x01\n\tScheduler\x12.\n\rScheduleBuild\x12\r.BuildRequest\x1a\x0c.BuildUpdate0\x01\x12\x38\n\x0b\x43\x61ncelBuild\x12\x13.CancelBuildRequest\x1a\x14.CancelBuildResponse\x12+\n\x0cScheduleTask\x12\x0c.TaskRequest\x1a\x0b.TaskUpdate0\x01\x42\x0eZ\x0cpkg/protocolb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'jolt.plugins.remote_execution.scheduler_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\014pkg/protocol'
  _BUILDREQUEST._serialized_start=68
  _BUILDREQUEST._serialized_end=159
  _BUILDUPDATE._serialized_start=161
  _BUILDUPDATE._serialized_end=222
  _CANCELBUILDREQUEST._serialized_start=224
  _CANCELBUILDREQUEST._serialized_end=262
  _CANCELBUILDRESPONSE._serialized_start=264
  _CANCELBUILDRESPONSE._serialized_end=315
  _TASKREQUEST._serialized_start=317
  _TASKREQUEST._serialized_end=365
  _WORKERALLOCATION._serialized_start=367
  _WORKERALLOCATION._serialized_end=415
  _TASKUPDATE._serialized_start=418
  _TASKUPDATE._serialized_end=581
  _SCHEDULER._serialized_start=584
  _SCHEDULER._serialized_end=746
# @@protoc_insertion_point(module_scope)



from __future__ import print_function
import glob
import os
import re
import sys
import time
import tqdm
if os.name == "nt":
    # FIXME: Workaround to make tqdm behave correctly on Windows
    import colorama
    colorama.deinit()  # Undo the work of tqdm
    os.system("")      # Hack to enable vt100
import traceback
from datetime import datetime
import threading
import logging
import logging.handlers
from contextlib import contextmanager
try:
    from StringIO import StringIO
except Exception:
    from io import StringIO

from jolt import config
from jolt.error import JoltError
from jolt import filesystem as fs
from jolt import colors
from jolt import common_pb2 as common_pb


current_time = datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%f")
logpath = config.get_logpath()
logfile = fs.path.join(logpath, f"{current_time}.log")
logcount = config.getint("jolt", "logcount", os.environ.get("JOLT_LOGCOUNT", 100))
logfiles = list(sorted(glob.glob(os.path.join(logpath, "*T*.log"))))

dirpath = fs.path.dirname(logfile)
if not fs.path.exists(dirpath):
    fs.makedirs(dirpath)

################################################################################

ERROR = common_pb.LogLevel.ERROR + 1
WARNING = common_pb.LogLevel.WARNING + 1
INFO = common_pb.LogLevel.INFO + 1
VERBOSE = common_pb.LogLevel.VERBOSE + 1
DEBUG = common_pb.LogLevel.DEBUG + 1
EXCEPTION = common_pb.LogLevel.EXCEPTION + 1
STDOUT = common_pb.LogLevel.STDOUT + 1
STDERR = common_pb.LogLevel.STDERR + 1
SILENCE = STDERR + 2

logging.addLevelName(VERBOSE, "VERBOSE")
logging.addLevelName(STDOUT, "STDOUT")
logging.addLevelName(STDERR, "STDERR")
logging.addLevelName(EXCEPTION, "EXCEPT")
logging.addLevelName(WARNING, "WARNING")
logging.addLevelName(INFO, "INFO")
logging.addLevelName(ERROR, "ERROR")
logging.addLevelName(DEBUG, "DEBUG")

logging.raiseExceptions = False


class Formatter(logging.Formatter):
    def __init__(self, fmt, *args, **kwargs):
        super(Formatter, self).__init__(*args, **kwargs)
        self.fmt = fmt

    def format(self, record):
        try:
            record.message = record.msg.format(*record.args)
        except Exception:
            record.message = record.msg
        record.asctime = datetime.fromtimestamp(record.created).strftime("%Y-%m-%d %H:%M:%S.%f")
        return self.fmt.format(
            levelname=record.levelname,
            message=record.message,
            asctime=record.asctime
        )


class ConsoleFormatter(logging.Formatter):
    def __init__(self, fmt_prefix, fmt_noprefix, *args, **kwargs):
        super(ConsoleFormatter, self).__init__(*args, **kwargs)
        self.fmt_prefix = fmt_prefix
        self.fmt_noprefix = fmt_noprefix
        self.always_prefix = False

    def enable_prefixes(self):
        self.always_prefix = True

    def enable_gdb(self):
        self.always_prefix = True
        self.fmt_prefix = "~\"" + self.fmt_prefix + "\\n\""

    def format(self, record):
        try:
            msg = record.msg.format(*record.args)
        except Exception:
            msg = record.msg
        if sys.stdout.isatty() and sys.stderr.isatty():
            if record.levelno >= ERROR:
                msg = colors.red(msg)
            elif record.levelno >= WARNING:
                msg = colors.yellow(msg)
        record.message = msg
        record.asctime = datetime.fromtimestamp(record.created).strftime("%Y-%m-%d %H:%M:%S.%f")
        record.prefix = True if record.__dict__.get("prefix", False) else False

        if not record.prefix and \
           not self.always_prefix and \
           record.levelno in [STDOUT, STDERR]:
            fmt = self.fmt_noprefix
        else:
            fmt = self.fmt_prefix

        return fmt.format(
            levelname=record.levelname,
            message=record.message,
            asctime=record.asctime
        )


class Filter(logging.Filter):
    def __init__(self, filterfn):
        self.filterfn = filterfn

    def filter(self, record):
        return self.filterfn(record)


class TqdmStream(object):
    def __init__(self, stream):
        self.stream = stream

    def write(self, msg):
        with tqdm.tqdm.external_write_mode(file=self.stream, nolock=False):
            self.stream.write(msg)

    def flush(self):
        getattr(self.stream, 'flush', lambda: None)()


# silence root logger
_root = logging.getLogger()
_root.setLevel(logging.CRITICAL)

# create jolt logger
_logger = logging.getLogger('jolt')
_logger.setLevel(EXCEPTION)

_console_formatter = ConsoleFormatter('[{levelname:>7}] {message}', '{message}')

if sys.stdout.isatty() and sys.stderr.isatty():
    _stdout = logging.StreamHandler(TqdmStream(sys.stdout))
else:
    _stdout = logging.StreamHandler(sys.stdout)
_stdout.setFormatter(_console_formatter)
_stdout.addFilter(Filter(lambda r: r.levelno < ERROR))
_stdout.addFilter(Filter(lambda r: r.levelno != EXCEPTION))

if sys.stdout.isatty() and sys.stderr.isatty():
    _stderr = logging.StreamHandler(TqdmStream(sys.stdout))
else:
    _stderr = logging.StreamHandler(sys.stderr)
_stderr.setFormatter(_console_formatter)
_stderr.addFilter(Filter(lambda r: r.levelno >= ERROR or r.levelno == EXCEPTION))

_logger.addHandler(_stdout)
_logger.addHandler(_stderr)

_file_formatter = Formatter('{asctime} [{levelname:>7}] {message}')


def start_file_log():
    global logfiles

    if len(logfiles) >= logcount:
        outdated = logfiles[:len(logfiles) - logcount + 1]
        logfiles = logfiles[-logcount + 1:]
        for file in outdated:
            os.unlink(file)

    _file = logging.FileHandler(logfile)
    _file.setLevel(EXCEPTION)
    _file.setFormatter(_file_formatter)
    _logger.addHandler(_file)


def log(level, message, created=None, context=None, prefix=False):
    created = created or time.time()
    message = f"[{context}] {message}" if context else message
    record = logging.LogRecord(
        name="log",
        level=level,
        pathname=__file__,
        lineno=0,
        msg=message,
        args={},
        exc_info=None,
    )
    record.created = created
    record.prefix = prefix
    _logger.handle(record)


def info(fmt, *args, **kwargs):
    _logger.log(INFO, fmt, *args, **kwargs)


def warning(fmt, *args, **kwargs):
    _logger.log(WARNING, fmt, *args, **kwargs)


def verbose(fmt, *args, **kwargs):
    _logger.log(VERBOSE, fmt, *args, **kwargs)


def debug(fmt, *args, **kwargs):
    _logger.log(DEBUG, fmt, *args, **kwargs)


def error(fmt, *args, **kwargs):
    _logger.log(ERROR, fmt, *args, **kwargs)


def stdout(line, **kwargs):
    line = line.replace("{", "{{")
    line = line.replace("}", "}}")
    _logger.log(STDOUT, line, extra=kwargs)


def stderr(line, **kwargs):
    line = line.replace("{", "{{")
    line = line.replace("}", "}}")
    _logger.log(STDERR, line, extra=kwargs)


def format_exception_msg(exc):
    te = traceback.TracebackException.from_exception(exc)

    if isinstance(exc, JoltError):
        return str(exc)

    elif isinstance(exc, SyntaxError):
        filename = fs.path.relpath(
            te.filename,
            fs.path.commonprefix([os.getcwd(), te.filename]))
        return "SyntaxError: {} ({}, line {})".format(
            te.text.strip(),
            filename,
            te.lineno)

    else:
        filename = fs.path.relpath(
            te.stack[-1].filename,
            fs.path.commonprefix([os.getcwd(), te.stack[-1].filename]))
        return "{}: {} ({}, line {}, in {})".format(
            type(exc).__name__,
            str(exc) or te.stack[-1].line,
            filename,
            te.stack[-1].lineno,
            te.stack[-1].name)


def exception(exc=None, error=True):
    if exc:
        if error:
            _logger.log(ERROR, format_exception_msg(exc))

        tb = traceback.format_exception(type(exc), value=exc, tb=exc.__traceback__)
        installdir = fs.path.dirname(__file__)
        if any(map(lambda frame: installdir not in frame, tb[1:-1])):
            while len(tb) > 2 and installdir in tb[1]:
                del tb[1]
        backtrace = "".join(tb).splitlines()
    else:
        backtrace = traceback.format_exc().splitlines()

    for line in backtrace:
        line = line.replace("{", "{{")
        line = line.replace("}", "}}")
        line = line.strip()
        _logger.log(EXCEPTION, line)


def transfer(line, context):
    context = "[{}] ".format(context)
    outline1 = context + line[10:]
    outline2 = context + line
    if line.startswith("[  ERROR]"):
        error(outline1)
    elif line.startswith("[WARNING]"):
        warning(outline1)
    elif line.startswith("[VERBOSE]"):
        verbose(outline1)
    elif line.startswith("[  DEBUG]"):
        debug(outline1)
    elif line.startswith("[   INFO]"):
        info(outline1)
    elif line.startswith("[ EXCEPT]"):
        outline1 = outline1.replace("{", "{{")
        outline1 = outline1.replace("}", "}}")
        _logger.log(EXCEPTION, outline1)
    elif line.startswith("[ STDERR]"):
        stderr(outline1, prefix=True)
    elif line.startswith("[ STDOUT]"):
        stdout(outline1, prefix=True)
    else:
        stdout(outline2, prefix=True)


def decompose(line):
    match = re.match(r"(?P<timestamp>[0-9]{4}-[0-9]{1,2}-[0-9]{1,2} [0-9]{1,2}:[0-9]{1,2}:[0-9]{1,2}.[0-9]{6}) \[(?P<loglevel>.*?)\]( \[(?P<context>.*?)\])? (?P<message>.*)", line)
    if not match:
        return line
    match = match.groupdict()
    return match["timestamp"], match["loglevel"].strip(), match.get("context"), match["message"]


class _Progress(object):
    def __init__(self, msg):
        verbose(msg)

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        pass

    def update(self, *args, **kwargs):
        pass


def progress_log(desc, count, unit):
    return _Progress(desc)


def progress(desc, count, unit, estimates=True, debug=False):
    bar_format = '{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}]' \
                 if not estimates else None
    if not debug and is_interactive() and not is_verbose():
        p = tqdm.tqdm(total=count, unit=unit, unit_scale=True, bar_format=bar_format, dynamic_ncols=True)
        p.set_description("[   INFO] " + desc)
        return p
    return progress_log(desc, count, unit)


_level = INFO


def level_to_pb(level):
    """ Convert a log level to a protobuf enum value. """
    return level - 1


def pb_to_level(level):
    """ Convert a protobuf enum value to a log level. """
    return level + 1


def set_level(level):
    """ Set the log level for terminal output. """

    if level not in [
        DEBUG,
        ERROR,
        EXCEPTION,
        INFO,
        SILENCE,
        STDERR,
        STDOUT,
        VERBOSE,
        WARNING,
    ]:
        raise ValueError("invalid log level")

    global _level
    _level = level
    _stdout.setLevel(level)
    _stderr.setLevel(level)


def set_level_pb(level):
    """ Set the log level from a protobuf enum value. """
    set_level(pb_to_level(level))


def get_level():
    global _level
    return _level


def get_level_pb():
    """ Get the log level as a protobuf enum value. """
    return level_to_pb(_level)


def set_worker():
    _console_formatter.enable_prefixes()


def enable_gdb():
    set_interactive(False)
    _console_formatter.enable_gdb()


def is_verbose():
    return _stdout.level <= VERBOSE


_interactive = True


def is_interactive():
    global _interactive
    return _interactive and sys.stdout.isatty() and sys.stderr.isatty()


def set_interactive(value):
    global _interactive
    _interactive = value


class _ThreadMapper(Filter):
    def __init__(self):
        self.thread_map = {}

    def map(self, fr, to):
        self.thread_map[fr] = to

    def unmap(self, fr):
        del self.thread_map[fr]

    def filter(self, record):
        record.thread = self.thread_map.get(record.thread, record.thread)
        return True


_thread_map = _ThreadMapper()


@contextmanager
def threadsink(level=DEBUG):
    threadid = threading.get_ident()
    stringbuf = StringIO()
    handler = logging.StreamHandler(stringbuf)
    handler.setLevel(level)
    handler.setFormatter(_file_formatter)
    handler.addFilter(_thread_map)
    handler.addFilter(Filter(lambda record: record.thread == threadid))
    _logger.addHandler(handler)
    try:
        yield stringbuf
    finally:
        _logger.removeHandler(handler)


@contextmanager
def handler(h):
    _logger.addHandler(h)
    try:
        yield
    finally:
        _logger.removeHandler(h)


@contextmanager
def map_thread(thread_from, thread_to):
    tid = thread_from.ident
    _thread_map.map(tid, thread_to.ident)
    try:
        yield
    finally:
        _thread_map.unmap(tid)


class _LogStream(object):
    def __init__(self):
        self.buf = ""

    def write(self, data):
        self.buf += data
        lines = self.buf.splitlines()
        if data[-1] != "\n":
            self.buf = lines[-1]
            lines = lines[:-1]
        else:
            self.buf = ""
        for line in lines:
            stdout(line)

    def flush(self):
        line = self.buf
        self.buf = ""
        if line:
            stdout(line)


@contextmanager
def stream():
    yield _LogStream()


set_level(STDOUT)



from jolt import attributes as jolt_attributes
from jolt import Download
from jolt import Parameter
from jolt.tasks import TaskRegistry

import platform


@jolt_attributes.system
@jolt_attributes.attribute("url", "url_{system}")
class Golang(Download):
    name = "golang"
    version = Parameter("1.18.3", help="Go version.")
    url_linux = "https://go.dev/dl/go{version}.linux-{arch}.tar.gz"
    url_windows = "https://go.dev/dl/go{version}.windows-{arch}.zip"

    @property
    def arch(self):
        arch = platform.machine()
        if arch == "x86_64":
            return "amd64"
        if arch in ["i386", "i486", "i586", "i686"]:
            return "i386"
        if arch in ["arm", "armv6l", "armv7l"]:
            return "armv6l"
        if arch in ["arm64", "armv8l", "aarch64"]:
            return "arm64"
        return None

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.PATH.append("go/bin")
        artifact.paths.goroot = "go"

    def unpack(self, artifact, tools):
        artifact.environ.GOROOT = str(artifact.paths.goroot)


TaskRegistry.get().add_task_class(Golang)



from datetime import datetime
import os

from jolt import utils
from jolt.hooks import TaskHook, TaskHookFactory
from jolt.config import get_workdir


class TimelineHooks(TaskHook):
    def __init__(self):
        self.tasks = []
        self.task_ids = {}

    def started(self, task):
        task._timeline_started = datetime.now().isoformat()
        self.task_ids[task] = len(self.tasks)

    def finished(self, task):
        task._timeline_finished = datetime.now().isoformat()
        self.tasks.append(task)
        self.render()

    def render(self):
        timeline = utils.render(
            "timeline.html.template",
            deps=self.deps,
            enumerate=enumerate,
            tasks=self.tasks)

        with open(os.path.join(get_workdir(), "timeline.html"), "w") as f:
            f.write(timeline)

    def deps(self, task):
        ids = []
        for anc in task.descendants:
            if anc in self.task_ids:
                ids.append(str(self.task_ids[anc]))
        return "null" if not ids else "'" + ",".join(ids) + "'"

    def task_started_download(self, task):
        self.started(task)

    def task_started_execution(self, task):
        self.started(task)

    def task_started_upload(self, task):
        self.started(task)

    def task_finished_download(self, task):
        self.finished(task)

    def task_finished_upload(self, task):
        self.finished(task)

    def task_finished_execution(self, task):
        self.finished(task)


@TaskHookFactory.register
class TimelineFactory(TaskHookFactory):
    def create(self, env):
        return TimelineHooks()



from os import _exit, environ, getcwd
import atexit
import click
import datetime
import platform
import subprocess
import sys
import uuid
import webbrowser

from jolt.tasks import Task, TaskRegistry, Parameter
from jolt import scheduler
from jolt import graph
from jolt import cache
from jolt import colors
from jolt import filesystem as fs
from jolt import log
from jolt import __version__
from jolt.log import logfile
from jolt import config
from jolt.loader import JoltLoader
from jolt import tools
from jolt import utils
from jolt.influence import HashInfluenceRegistry
from jolt.options import JoltOptions
from jolt import hooks
from jolt.manifest import JoltManifest
from jolt.error import JoltError
from jolt.error import raise_error
from jolt.error import raise_error_if
from jolt.error import raise_task_error_if
from jolt.plugins import report

debug_enabled = False
workdir = getcwd()


class ArgRequiredUnless(click.Argument):
    def __init__(self, *args, **kwargs):
        self.required_unless = kwargs.pop('required_unless')
        assert self.required_unless, "'required_unless' parameter required"
        super(ArgRequiredUnless, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.required_unless not in opts:
            if not opts.get(self.name, None):
                self.required = True
        return super(ArgRequiredUnless, self).handle_parse_result(
            ctx, opts, args)


class PluginGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        if cmd_name == "info":
            cmd_name = "inspect"

        if cmd_name in ["export", "inspect"]:
            log.set_level(log.SILENCE)
        elif ctx.params.get("verbose") >= 3:
            log.set_level(log.EXCEPTION)
        elif ctx.params.get("verbose") >= 2:
            log.set_level(log.DEBUG)
        elif ctx.params.get("verbose") >= 1:
            log.set_level(log.VERBOSE)

        config_files = ctx.params.get("config_file") or []
        for config_file in config_files:
            log.verbose("Config: {0}", config_file)
            config.load_or_set(config_file)

        # Load configured plugins
        JoltLoader.get().load_plugins()

        return click.Group.get_command(self, ctx, cmd_name)


@click.group(cls=PluginGroup, invoke_without_command=True)
@click.version_option(__version__)
@click.option("-v", "--verbose", count=True, help="Verbose output (repeat to raise verbosity).")
@click.option("-c", "--config", "config_file", multiple=True, type=str,
              help="Load a configuration file or set a configuration key.")
@click.option("-d", "--debugger", is_flag=True,
              help="Attach debugger on exception.")
@click.option("-p", "--profile", is_flag=True, hidden=True,
              help="Profile code while running")
@click.option("-f", "--force", is_flag=True, default=False, hidden=True,
              help="Force rebuild of target tasks.")
@click.option("-s", "--salt", type=str, hidden=True,
              help="Add salt as task influence.")
@click.option("-g", "--debug", is_flag=True, default=False, hidden=True,
              help="Start debug shell before executing task.")
@click.option("-n", "--network", is_flag=True, default=False, hidden=True,
              help="Build on network.")
@click.option("-l", "--local", is_flag=True, default=False, hidden=True,
              help="Disable all network operations.")
@click.option("-k", "--keep-going", is_flag=True, default=False, hidden=True,
              help="Build as many tasks as possible, don't abort on first failure.")
@click.option("-j", "--jobs", type=int, default=1, hidden=True,
              help="Number of tasks allowed to execute in parallel (1). ")
@click.option("-h", "--help", is_flag=True, help="Show this message and exit.")
@click.pass_context
def cli(ctx, verbose, config_file, debugger, profile,
        force, salt, debug, network, local, keep_going, jobs, help):
    """
    A task execution tool.

    When invoked without any commands and arguments, Jolt by default tries
    to execute and build the artifact of a task called `default`. To build
    artifacts of other tasks use the build subcommand.

    The Jolt command line interface is hierarchical. One set of options
    can be passed to the top-level command and a different set of options
    to the subcommands, simultaneously. For example, verbose output is
    a top-level option while forced rebuild is a build command option.
    They may combined like this:

      $ jolt --verbose build --force taskname

    Most build command options are available also at the top-level when
    build is invoked implicitly for the default task.

    """

    global debug_enabled
    debug_enabled = debugger

    if ctx.invoked_subcommand not in ["log", "report"]:
        log.start_file_log()

    log.verbose("Jolt command: {}", " ".join([fs.path.basename(sys.argv[0])] + sys.argv[1:]))
    log.verbose("Jolt host: {}", environ.get("HOSTNAME", "localhost"))
    log.verbose("Jolt install path: {}", fs.path.dirname(__file__))

    if ctx.invoked_subcommand in ["config", "log"]:
        # Don't attempt to load any task recipes as they might require
        # plugins that are not yet configured.
        return

    if ctx.invoked_subcommand is None:
        build = ctx.command.get_command(ctx, "build")

    if help:
        print(ctx.get_help())
        sys.exit(0)

    manifest = JoltManifest()
    utils.call_and_catch(manifest.parse)
    manifest.process_import()
    ctx.obj["manifest"] = manifest

    if manifest.version:
        from jolt.version_utils import requirement, version
        req = requirement(manifest.version)
        ver = version(__version__)
        raise_error_if(not req.satisfied(ver),
                       "This project requires Jolt version {} (running {})",
                       req, __version__)

    loader = JoltLoader.get()
    tasks = loader.load()
    for cls in tasks:
        TaskRegistry.get().add_task_class(cls)

    if ctx.invoked_subcommand in ["build", "clean"] and loader.joltdir:
        ctx.obj["workspace_lock"] = utils.LockFile(
            fs.path.join(loader.joltdir, "build"),
            log.info, "Workspace is locked by another process, please wait...")
        atexit.register(ctx.obj["workspace_lock"].close)

    # If no command is given, we default to building the default task.
    # If the default task doesn't exist, help is printed inside build().
    if ctx.invoked_subcommand is None:
        task = config.get("jolt", "default", "default")
        taskname, _ = utils.parse_task_name(task)
        if TaskRegistry.get().get_task_class(taskname) is not None:
            ctx.invoke(build, task=[task], force=force, salt=salt, debug=debug,
                       network=network, local=local, keep_going=keep_going, jobs=jobs)
        else:
            print(cli.get_help(ctx))
            sys.exit(1)


def _autocomplete_tasks(ctx, args, incomplete):
    manifest = JoltManifest()
    utils.call_and_catch(manifest.parse)
    manifest.process_import()

    tasks = JoltLoader.get().load()
    tasks = [task.name for task in tasks if task.name.startswith(incomplete or '')]
    return sorted(tasks)


@cli.command()
@click.argument("task", type=str, nargs=-1, shell_complete=_autocomplete_tasks, cls=ArgRequiredUnless, required_unless="worker")
@click.option("-c", "--copy", type=click.Path(),
              help="Copy artifact content to directory PATH.")
@click.option("-d", "--default", type=str, multiple=True, help="Override default parameter values.", metavar="DEFAULT")
@click.option("-f", "--force", is_flag=True, default=False, help="Force rebuild of TASK artifact.")
@click.option("-g", "--debug", is_flag=True, default=False,
              help="Start debug shell before executing TASK.")
@click.option("-j", "--jobs", type=int, default=1, help="Number of tasks allowed to execute in parallel [1]. ", metavar="JOBS")
@click.option("-k", "--keep-going", is_flag=True, default=False, help="Build as many task artifacts as possible.")
@click.option("-l", "--local", is_flag=True, default=False, help="Disable remote cache access.")
@click.option("-n", "--network", is_flag=True, default=False, help="Distribute tasks to network workers.")
@click.option("-s", "--salt", type=str, help="Add salt as hash influence for all tasks in dependency tree.", metavar="SALT")
@click.option("-m", "--mute", is_flag=True, help="Display task log only if it fails.")
@click.option("-v", "--verbose", count=True, help="Verbose output (repeat to raise verbosity).")
@click.option("--result", type=click.Path(), hidden=True,
              help="Write result manifest to this file.")
@click.option("--no-download", is_flag=True, default=False,
              help="Don't download artifacts from remote storage")
@click.option("--no-upload", is_flag=True, default=False,
              help="Don't upload artifacts to remote storage")
@click.option("--download", is_flag=True, default=False,
              help="Do download artifacts from remote storage")
@click.option("--upload", is_flag=True, default=False,
              help="Do upload artifacts to remote storage")
@click.option("--no-prune", is_flag=True, default=False,
              help="Don't prune cached artifacts from the build graph. This option can be used to populate the local cache with remotely cached dependency artifacts.")
@click.option("--worker", is_flag=True, default=False,
              help="Run with the worker build strategy", hidden=True)
@click.pass_context
@hooks.cli_build
def build(ctx, task, network, keep_going, default, local,
          no_download, no_upload, download, upload, worker, force,
          salt, copy, debug, result, jobs, no_prune, verbose,
          mute):
    """
    Build task artifact.

    TASK is the name of the task to execute. It is optionally followed by a colon and
    parameter value assignments. Assignments are separated by commas. Example:

       taskname:param1=value1,param2=value2

    Default parameter values can be overridden for any task in the dependency tree
    with --default. DEFAULT is a qualified task name, just like TASK, but parameter
    assignments change default values.

    By default, a task is executed locally and the resulting artifact is stored
    in the local artifact cache. If an artifact is already available in the cache,
    no execution takes place. Artifacts are identified with a hash digest,
    constructed from hashing task attributes.

    When remote cache providers are configured, artifacts may be downloaded from and/or
    uploaded to the remote cache as execution progresses. Several options exist to control
    the behavior, such as --local which disables all remote caches.

    Distributed task execution is enabled by passing the --network option. Tasks are then
    distributed to and executed by a pool of workers, if one has been configured.

    Rebuilds can be forced with either --force or --salt. --force rebuilds the requested
    task, but not its dependencies. --salt affects the entire dependency tree. Both add
    an extra attribute to the task hash calculation in order to taint the identity and
    induce a cache miss. In both cases, existing intermediate files in build directories
    are removed before execution starts.

    """
    raise_error_if(network and local,
                   "The -n and -l flags are mutually exclusive")

    raise_error_if(network and debug,
                   "The -g and -n flags are mutually exclusive")

    raise_error_if(no_download and download,
                   "The --download and --no-download flags are mutually exclusive")

    raise_error_if(no_upload and upload,
                   "The --upload and --no-upload flags are mutually exclusive")

    if verbose >= 3:
        log.set_level(log.EXCEPTION)
    elif verbose >= 2:
        log.set_level(log.DEBUG)
    elif verbose >= 1:
        log.set_level(log.VERBOSE)

    ts_start = utils.duration()
    task = list(task)
    task = [utils.stable_task_name(t) for t in task]

    if network:
        _download = config.getboolean("network", "download", True)
        _upload = config.getboolean("network", "upload", True)
    else:
        _download = config.getboolean("jolt", "download", True)
        _upload = config.getboolean("jolt", "upload", True)

    if local:
        _download = False
        _upload = False
    else:
        if no_download:
            _download = False
        if no_upload:
            _upload = False
        if download:
            _download = True
        if upload:
            _upload = True

    options = JoltOptions(
        network=network,
        local=local,
        download=_download,
        upload=_upload,
        keep_going=keep_going,
        default=default,
        worker=worker,
        debug=debug,
        salt=salt,
        jobs=jobs,
        mute=mute)

    acache = cache.ArtifactCache.get(options)

    executors = scheduler.ExecutorRegistry.get(options)
    if worker:
        log.set_worker()
        log.verbose("Local build as a worker")
        strategy = scheduler.WorkerStrategy(executors, acache)
    elif network:
        log.verbose("Distributed build as a user")
        strategy = scheduler.DistributedStrategy(executors, acache)
    else:
        log.verbose("Local build as a user")
        strategy = scheduler.LocalStrategy(executors, acache)

    hooks.TaskHookRegistry.get(options)
    registry = TaskRegistry.get(options)

    for params in default:
        registry.set_default_parameters(params)

    manifest = ctx.obj["manifest"]

    for mb in manifest.builds:
        for mt in mb.tasks:
            task.append(mt.name)
        for mt in mb.defaults:
            registry.set_default_parameters(mt.name)

    if force:
        for goal in task:
            registry.get_task(goal, manifest=manifest).taint = uuid.uuid4()

    log.info("Started: {}", datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

    gb = graph.GraphBuilder(registry, acache, manifest, options, progress=True)
    dag = gb.build(task)

    # Collect information about artifact presence before starting prune or build
    acache.precheck(dag.persistent_artifacts, remote=not local)

    if not no_prune:
        gp = graph.GraphPruner(acache, strategy)
        dag = gp.prune(dag)

    goal_tasks = dag.goals
    goal_task_duration = 0

    session = executors.create_session(dag) if options.network else {}
    queue = scheduler.TaskQueue(strategy, acache, session)

    try:
        if not dag.has_tasks():
            return

        progress = log.progress(
            "Progress",
            dag.number_of_tasks(filterfn=lambda t: not t.is_resource()),
            " tasks",
            estimates=False,
            debug=debug)

        with progress:
            while dag.has_tasks() or not queue.empty():
                # Find all tasks ready to be executed
                leafs = dag.select(lambda graph, task: task.is_ready())

                # Order the tasks by their weights to improve build times
                leafs.sort(key=lambda x: x.weight)

                while leafs:
                    task = leafs.pop()
                    queue.submit(task)

                task, error = queue.wait()

                if not task:
                    dag.debug()
                    break
                elif task.is_goal() and task.duration_running:
                    goal_task_duration += task.duration_running.seconds

                if not task.is_resource():
                    if no_prune and task.task.unpack.__func__ is not Task.unpack:
                        with acache.get_context(task):
                            pass

                    progress.update(1)

                if not keep_going and error is not None:
                    queue.abort()
                    task.raise_for_status()
                    raise error

        if dag.failed or dag.unstable:
            log.error("List of failed tasks")
            for failed in dag.failed + dag.unstable:
                log.error("- {}", failed.log_name.strip("()"))

        for failed_task in dag.failed:
            failed_task.raise_for_status()
        if dag.failed:
            raise_error("No more tasks could be executed")

    except KeyboardInterrupt:
        print()
        log.warning("Interrupted by user")
        try:
            queue.abort()
            sys.exit(1)
        except KeyboardInterrupt:
            print()
            log.warning("Interrupted again, exiting")
            _exit(1)
    finally:
        queue.shutdown()

        for task in goal_tasks:
            for artifact in task.artifacts:
                if acache.is_available_locally(artifact):
                    log.info("Location: {0}", artifact.path)
                    if copy:
                        dst = utils.as_dirpath(fs.path.join(workdir, click.format_filename(copy)))
                        artifact.copy("*", dst, symlinks=True)

        log.info("Ended: {}", datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
        log.info("Total execution time: {0} {1}",
                 str(ts_start),
                 str(queue.duration_acc) if network else '')
        if result:
            with report.update() as manifest:
                manifest.duration = str(goal_task_duration)
                manifest.write(result)


@cli.command()
@click.argument("task", type=str, nargs=-1, required=False, shell_complete=_autocomplete_tasks)
@click.option("-d", "--deps", is_flag=True, help="Clean all task dependencies.")
@click.option("-e", "--expired", is_flag=True, help="Only clean expired tasks.")
@click.pass_context
@hooks.cli_clean
def clean(ctx, task, deps, expired):
    """
    Delete task artifacts and intermediate files.

    When run without arguments, this command removes all task artifacts
    from the local cache, but no intermediate files are removed.

    When TASK is specified, the task clean() method is invoked to remove
    any intermediate files still present in persistent build directories.
    Secondly, the task artifact will be removed from the local cache.
    Global caches are not affected. The --deps parameter can be used to also
    clean all dependencies of the specified TASK.

    By default, task artifacts are removed without considering any
    artifact expiration metadata. To only remove artifact which have expired,
    use the --expired parameter. Artifacts typically expire immediately after
    creation unless explicitly configured not to.
    """
    acache = cache.ArtifactCache.get()
    if task:
        task = [utils.stable_task_name(t) for t in task]
        registry = TaskRegistry.get()
        dag = graph.GraphBuilder(registry, acache, ctx.obj["manifest"]).build(task)
        if deps:
            tasks = dag.tasks
        else:
            tasks = dag.goals
        for task in tasks:
            task.clean(acache, expired)
    else:
        acache.discard_all(expired, onerror=fs.onerror_warning)
        try:
            # May not be in a workspace when running the command
            t = tools.Tools()
            fs.rmtree(t.buildroot, onerror=fs.onerror_warning)
        except AssertionError:
            pass


@cli.command(name="config")
@click.option("-l", "--list", is_flag=True,
              help="List all configuration keys and values.")
@click.option("-d", "--delete", is_flag=True,
              help="Delete configuration key.")
@click.option("-g", "--global", "global_", is_flag=True,
              help="List, set or get configuration keys in the global config.")
@click.option("-u", "--user", is_flag=True,
              help="List, set or get configuration keys in the user config.")
@click.argument("key", type=str, nargs=1, required=False)
@click.argument("value", type=str, nargs=1, required=False)
@click.pass_context
def _config(ctx, list, delete, global_, user, key, value):
    """
    Configure Jolt.

    You can query/set/replace/unset configuration keys with this command.
    Key strings are constructed from the configuration section and the
    option separated by a dot.

    There are tree different configuration sources:

       - A global configuration file

       - A user configuration file

       - Temporary configuration passed on the command line.

    When reading, the values are read from all configuration sources.
    The options --global and --user can be used to tell the command to read
    from only one of the sources. If a configuration key is available from
    multiple sources, temporary CLI configuration has priority followed by
    the user configuration file and lastly the global configuration file.

    When writing, the new values are written to the user configuration by default.
    The options --global and --user can be used to tell the command to write
    to only one of the sources.

    When removing keys, the values are removed from all sources.
    The options --global and --user can be used to restrict removal to one of
    the sources.

    To assign a value to a key:

      $ jolt config jolt.default all   # Change name of the default task

    To list existing keys:

      $ jolt config -l                 # List all existing keys

      $ jolt config -l -g              # List keys in the global config file

      $ jolt config jolt.colors        # Display the value of a key.

    To delete an existing key:

      $ jolt config -d jolt.colors

    To pass temporary configuration:

      $ jolt -c jolt.colors=true config -l

    """

    if delete and not key:
        raise click.UsageError("--delete requires KEY")

    if not key and not list and not key:
        print(ctx.get_help())
        sys.exit(1)

    if global_ and user:
        raise click.UsageError("--global and --user are mutually exclusive")

    alias = None

    if global_:
        alias = "global"
    if user:
        alias = "user"

    def _print_key(section, opt):
        value = config.get(section, opt, alias=alias)
        raise_error_if(value is None, "No such key: {}".format(key))
        print("{} = {}".format(key, value))

    def _print_section(section):
        print("{}".format(section))

    if list:
        for section, option, value in config.items(alias):
            if option:
                print("{}.{} = {}".format(section, option, value))
            else:
                print(section)
    elif delete:
        raise_error_if(config.delete(key, alias) <= 0,
                       "No such key: {}", key)
        config.save()
    elif key:
        section, opt = config.split(key)
        if value:
            raise_error_if(opt is None, "Invalid configuration key: {}".format(key))
            config.set(section, opt, value, alias)
            try:
                config.save()
            except Exception as e:
                log.exception()
                raise_error("Failed to write configuration file: {}".format(e))
        else:
            if opt:
                _print_key(section, opt)
            else:
                _print_section(section)


@cli.command()
@click.argument("task", type=str, nargs=-1, required=False, shell_complete=_autocomplete_tasks)
@click.option("-c", "--cached", "show_cache", is_flag=True, help="Highlight cache presence with colors.")
@click.option("-r", "--reverse", type=str, help="Display consumers of REVERSE if TASK is executed.")
@click.option("-p", "--prune", "prune", is_flag=True, help="Do not repeat tasks. An already visited task's name is printed inside [square brackets] and its deps are omitted.")
@click.pass_context
def display(ctx, task, reverse=None, show_cache=False, prune=False):
    """
    Display a task and its dependencies visually.

    """
    registry = TaskRegistry.get()
    options = JoltOptions()
    acache = cache.ArtifactCache.get(options)
    gb = graph.GraphBuilder(registry, acache, ctx.obj["manifest"])
    dag = gb.build(task, influence=show_cache)

    if reverse:
        def iterator(task):
            return list(dag.predecessors(task))
        reverse = utils.as_list(reverse)
        tasklist = dag.select(
            lambda graph, node:
            node.short_qualified_name in reverse or node.qualified_name in reverse)
    else:
        def iterator(task):
            return task.children
        tasklist = dag.requested_goals

    if dag.has_tasks():
        processed = set()

        def _display(task, indent=0, last=None):
            header = ""
            if indent > 0:
                for pipe in last[:-1]:
                    if pipe:
                        header += "\u2502 "
                    else:
                        header += "  "
                if last[-1]:
                    header += "\u251c\u2574"
                else:
                    header += "\u2514\u2574"

            if not show_cache:
                colorize = str
            elif task.is_cacheable() and not acache.is_available(task):
                colorize = colors.red
            else:
                colorize = colors.green

            if prune and task.short_qualified_name in processed:
                def prune_marker(n):
                    return "[" + n + "]"
            else:
                def prune_marker(n):
                    return n

            print(header + colorize(prune_marker(task.short_qualified_name)))
            children = iterator(task)
            if not prune or task.short_qualified_name not in processed:
                for i in range(0, len(children)):
                    _display(children[i], indent + 1, last=(last or []) + [i + 1 != len(children)])

            if prune:
                processed.add(task.short_qualified_name)

        for task in tasklist:
            _display(task)
    else:
        log.info("no tasks to display")


@cli.command()
def docs():
    """
    Opens the Jolt documentation in the default webbrowser.
    """
    webbrowser.open(config.get("jolt", "docs", "http://jolt.readthedocs.io/"))


@cli.command()
@click.argument("task", type=str, nargs=-1, required=True, shell_complete=_autocomplete_tasks)
@click.option("-c", "--copy", type=click.Path(),
              help="Copy artifact content to directory PATH.")
@click.option("-ca", "--copy-all", type=click.Path(), help="Copy artifacts, including dependency artifacts, to directory PATH. Implies --deps.")
@click.option("-d", "--deps", is_flag=True, help="Download dependencies.")
@click.pass_context
@hooks.cli_download
def download(ctx, task, deps, copy, copy_all):
    """
    Download task artifacts from remote caches.

    No attempt to build the artifact is made if it is not present
    in configured remote caches.

    Task dependencies may optionally be downloaded by passing the -d option.
    """

    raise_error_if(copy and copy_all, "--copy and --copy-all are mutually exclusive")
    if copy_all:
        deps = True

    manifest = ctx.obj["manifest"]
    options = JoltOptions()
    acache = cache.ArtifactCache.get(options)
    hooks.TaskHookRegistry.get(options)
    executors = scheduler.ExecutorRegistry.get(options)
    registry = TaskRegistry.get()
    strategy = scheduler.DownloadStrategy(executors, acache)
    queue = scheduler.TaskQueue(strategy, acache, {})
    gb = graph.GraphBuilder(registry, acache, manifest, options, progress=True)
    dag = gb.build(task)

    if not deps:
        for task in dag.tasks:
            if not task.is_goal():
                task.pruned()

    all_tasks = dag.tasks
    goal_tasks = dag.goals

    try:
        with log.progress("Progress", dag.number_of_tasks(), " tasks", estimates=False, debug=False) as p:
            while dag.has_tasks() or not queue.empty():
                leafs = dag.select(lambda graph, task: task.is_ready())

                while leafs:
                    task = leafs.pop()
                    queue.submit(task)

                task, error = queue.wait()
                p.update(1)

        copy_tasks = goal_tasks if not copy_all else all_tasks
        for goal in copy_tasks:
            if goal.is_available_locally():
                for artifact in goal.artifacts:
                    if copy:
                        log.info("Copying: {0}", artifact.path)
                        artifact.copy("*", utils.as_dirpath(fs.path.join(workdir, click.format_filename(copy))), symlinks=True)
                    elif copy_all:
                        log.info("Copying: {0}", artifact.path)
                        artifact.copy("*", utils.as_dirpath(fs.path.join(workdir, click.format_filename(copy_all))), symlinks=True)
                    elif goal.is_goal():
                        log.info("Location: {0}", artifact.path)

    except KeyboardInterrupt:
        print()
        log.warning("Interrupted by user")
        try:
            queue.abort()
            sys.exit(1)
        except KeyboardInterrupt:
            print()
            log.warning("Interrupted again, exiting")
            _exit(1)

    except Exception as e:
        log.set_interactive(True)
        raise e

    finally:
        queue.shutdown()


@cli.command(hidden=True)
@click.argument("task", type=str, nargs=-1, required=True)
@click.option("-r", "--remove", is_flag=True, help="Remove tasks from existing manifest.")
@click.option("-d", "--default", type=str, multiple=True, help="Override default parameter values.")
@click.option("-o", "--output", type=str, default="default.joltxmanifest", help="Manifest filename.")
@click.pass_context
def freeze(ctx, task, default, output, remove):
    """
    Freeze the identity of a task.

    <WIP>
    """
    manifest = ctx.obj["manifest"]

    options = JoltOptions(default=default)
    acache = cache.ArtifactCache.get(options)
    scheduler.ExecutorRegistry.get(options)
    registry = TaskRegistry.get()

    for params in default:
        registry.set_default_parameters(params)

    gb = graph.GraphBuilder(registry, acache, manifest)
    dag = gb.build(task)

    available, missing = acache.availability(dag.persistent_artifacts)

    for artifact in missing:
        raise_task_error_if(
            not remove, artifact.get_task(),
            "Task artifact is not available in any cache, build it first")

    for task in dag.tasks:
        if task.is_resource() or not task.is_cacheable():
            continue
        manifest_task = manifest.find_task(task)
        if remove and manifest_task:
            manifest.remove_task(manifest_task)
            continue
        if not remove:
            if not manifest_task:
                manifest_task = manifest.create_task()
            manifest_task.name = task.qualified_name
            manifest_task.identity = task.identity

    manifest.write(fs.path.join(JoltLoader.get().joltdir, output))


@cli.command(name="list")
@click.argument("task", type=str, nargs=-1, required=False, shell_complete=_autocomplete_tasks)
@click.option("-a", "--all", is_flag=True, help="List all direct and indirect dependencies of TASK.")
@click.option("-r", "--reverse", type=str, help="Only list dependencies of TASK that are also reverse dependencies of REVERSE.", metavar="REVERSE")
@click.pass_context
def _list(ctx, task=None, all=False, reverse=None):
    """
    List all tasks, or dependencies of a task.

    By default, when no TASK is specified, all known task names
    are listed in alphabetical order.

    When a TASK is specified, only direct dependencies of that task
    are listed. Use -a to also list its indirect dependencies.

    Multiple TASK names are allowed.
    """

    raise_error_if(not task and reverse, "TASK required with --reverse")

    options = JoltOptions()
    acache = cache.ArtifactCache.get(options)
    registry = TaskRegistry.get()

    if not task:
        classes = registry.get_task_classes()
        for task in sorted(classes, key=lambda x: x.name):
            if task.name:
                print(task.name)
        return

    task = [utils.stable_task_name(t) for t in task]
    reverse = [utils.stable_task_name(t) for t in utils.as_list(reverse or [])]

    try:
        dag = graph.GraphBuilder(registry, acache, ctx.obj["manifest"]).build(task, influence=False)
    except JoltError as e:
        raise e
    except Exception:
        raise_error("An exception occurred during task dependency evaluation, see log for details")

    task = reverse or task
    nodes = dag.select(
        lambda graph, node:
        node.short_qualified_name in task or node.qualified_name in task)
    nodes = list(nodes)
    iterator = dag.predecessors if reverse else dag.successors

    tasklist = set()
    while nodes:
        node = nodes.pop()
        for task in iterator(node):
            if all and task.short_qualified_name not in tasklist:
                new_node = dag.get_task(task.qualified_name)
                nodes.append(new_node)
            tasklist.add(task.short_qualified_name)

    for task in sorted(list(tasklist)):
        print(task)


@cli.command(name="log")
@click.option("-f", "--follow", is_flag=True, help="Display log output as it appears")
@click.option("-d", "--delete", is_flag=True, help="Delete the log file")
def _log(follow, delete):
    """
    Display the Jolt log file.

    """
    if not log.logfiles:
        print("No logs exist")
        return

    if follow:
        subprocess.call("tail -f {0}".format(log.logfiles[-1]), shell=True)
    elif delete:
        for file in log.logfiles:
            fs.unlink(file)
    else:
        t = tools.Tools()
        configured_pager = config.get("jolt", "pager", environ.get("PAGER", None))
        for pager in [configured_pager, "less", "more", "cat"]:
            if pager and t.which(pager):
                return subprocess.call("{1} {0}".format(log.logfiles[-1], pager), shell=True)
        print(t.read_file(logfile))


@cli.command()
@click.argument("task", shell_complete=_autocomplete_tasks)
@click.option("-i", "--influence", is_flag=True, help="Print influence attributes and values.")
@click.option("-a", "--artifact", is_flag=True, help="Print artifact cache status.")
@click.option("-s", "--salt", type=str, help="Add salt as task influence.", metavar="SALT")
@click.pass_context
def inspect(ctx, task, influence=False, artifact=False, salt=None):
    """
    View information about a task.

    This command displays information about a task, such as its class
    documentation, parameters and their accepted values, requirements,
    task class origin (file/line), influence attributes, artifact identity,
    cache status, and more. Default parameter values, if any, are highlighted.

    """
    task_name = task
    task_cls_name, task_params = utils.parse_task_name(task_name)
    task_registry = TaskRegistry.get()
    task = task_registry.get_task_class(task_cls_name)
    raise_task_error_if(not task, task_name, "No such task")

    from jolt import inspection

    print()
    print("  {0}".format(task.name))
    print()
    if task.__doc__:
        print("  {0}".format(task.__doc__.strip()))
        print()
    print("  Parameters")
    has_param = False
    params = {key: getattr(task, key) for key in dir(task)
              if isinstance(utils.getattr_safe(task, key), Parameter)}
    for item, param in params.items():
        has_param = True
        print("    {0:<15}   {1}".format(item, param.help or ""))
    if not has_param:
        print("    None")

    print()
    print("  Definition")
    print("    {0:<15}   {1} ({2})".format(
        "File", fs.path.relpath(inspection.getfile(task), JoltLoader.get().joltdir),
        inspection.getlineno(task)))

    print()
    print("  Requirements")
    manifest = ctx.obj["manifest"]
    try:
        task = task_registry.get_task(task_name, manifest=manifest)
        for req in sorted(utils.as_list(utils.call_or_return(task, task.requires))):
            print("    {0}".format(task.tools.expand(req)))
        if not task.requires:
            print("    None")
        print()
    except Exception as e:
        log.exception()
        if "has not been set" in str(e):
            print("    Unavailable (parameters must be set)")
            print()
            return
        print("    Unavailable (exception during evaluation)")
        print()
        return

    if salt:
        task.taint = salt

    if artifact:
        options = JoltOptions(salt=salt)
        acache = cache.ArtifactCache.get()
        builder = graph.GraphBuilder(task_registry, acache, manifest, options)
        dag = builder.build([task.qualified_name])
        tasks = dag.select(lambda graph, node: node.task is task)
        assert len(tasks) == 1, "graph produced multiple tasks, one expected"
        proxy = tasks[0]
        task = proxy.task

        print("  Cache")
        print("    Identity          {0}".format(proxy.identity))
        if proxy.is_available_locally():
            for artifact in filter(lambda a: not a.is_session(), proxy.artifacts):
                print("    Location          {0}".format(artifact.path))
            print("    Local             True ({0})".format(
                utils.as_human_size(sum([artifact.get_size() for artifact in proxy.artifacts]))))
        else:
            print("    Local             False")
        print("    Remote            {0}".format(proxy.is_available_remotely(cache=False)))
        print()

    if influence:
        print("  Influence")
        for string in HashInfluenceRegistry.get().get_strings(task):
            string = string.split(":", 1)
            print("    {:<18}{}".format(string[0][10:], string[1].strip()))


@cli.command()
@click.argument("task", type=str, nargs=-1, required=True, shell_complete=_autocomplete_tasks)
@click.pass_context
def export(ctx, task):
    """
    Export task artifact metadata into shell environment.

    When running the command, a shell script is printed to stdout.
    It can be redirected to a file and executed separately, or sourced
    directly in the shell by running:

      source <(jolt export <task>)

    The script creates a virtual environment that is identical
    to the environment that would be setup when the specified task
    artifact is consumed by another task. This enables packaged
    applications, such as compilers, to be run directly from the shell.

    The command will fail if the task artifact or any dependency
    artifact is missing in the local cache. Build the task to
    populate the cache and then try again.

    Run ``deactivate-jolt`` to leave the virtual environment. All
    environment variables will be restored to their original values.
    """
    try:
        _export(ctx, task)
    finally:
        log.set_level(log.INFO)


def _export(ctx, task):
    acache = cache.ArtifactCache.get()
    task = [utils.stable_task_name(t) for t in task]
    registry = TaskRegistry.get()
    executors = scheduler.ExecutorRegistry.get()
    strategy = scheduler.LocalStrategy(executors, acache)

    dag = graph.GraphBuilder(registry, acache, ctx.obj["manifest"])
    dag = dag.build(task)

    gp = graph.GraphPruner(acache, strategy)
    dag = gp.prune(dag)

    class Export(object):
        def __init__(self):
            self.environ = {}
            self.prepend_environ = {}

        def setenv(self, name, value):
            self.environ[name] = value

    class Context(object):
        def __init__(self, tasks):
            self.tasks = tasks
            self.environ = set()
            self.exports = {}

        def add_export(self, task, visitor):
            self.exports[task] = visitor
            self.environ.update(set(visitor.environ.keys()))

    tasks = list(filter(lambda t: t.is_cacheable(), reversed(dag.topological_nodes)))
    context = Context(tasks)

    for task in context.tasks:
        for artifact in task.artifacts:
            raise_task_error_if(
                artifact.is_temporary(), task,
                "Task artifact not found in local cache, build it first")

            visitor = Export()
            cache.visit_artifact(task, artifact, visitor)
            context.add_export(task, visitor)

    script = utils.render(
        "export.sh.template",
        ctx=context)

    print(script)


@cli.command(name="report")
@click.pass_context
@hooks.cli_report
def _report(ctx):
    """Create and publish a report with system information and logs.

    The purpose of the report command is to aid troubleshooting by
    providing users a method to quickly generate and share a report
    with information about the environment and previous activity.

    The report is uploaded to remote caches if configured, otherwise
    a report archive is created in the current working directory.
    """

    options = JoltOptions()
    acache = cache.ArtifactCache.get(options)

    with tools.Tools() as t:

        artifact = cache.Artifact(
            acache,
            None,
            identity=str(uuid.uuid4()),
            name="jolt",
            session=True,
            tools=t)

        with t.tmpdir("report") as tmp, t.cwd(tmp.path):
            log.info("Collecting environment")
            env = ""
            for key, val in environ.items():
                env += f"{key} = {val}\n"
            t.write_file("environ.txt", env, expand=False)

            log.info("Collecting configuration")
            config.save(tmp.path)
            artifact.collect("*.conf", "configs/")

            log.info("Collecting platform information")
            uname = platform.uname()
            libc = " ".join(platform.libc_ver())
            pyver = platform.python_version()
            pyimpl = platform.python_implementation()
            pfm = f"""System: {uname.system}
Node: {uname.node}
Release: {uname.release}
Version: {uname.version}
Machine: {uname.machine}
Libc: {libc}
Python: {pyimpl} {pyver}
"""
            t.write_file("platform.txt", pfm, expand=False)

            def collect_command(what, cmd):
                try:
                    log.info("Collecting {}", what)
                    t.run(cmd, output=False)
                except Exception:
                    pass

            collect_command("pip modules", "pip list > packages-pip.txt")
            if uname.system == "Linux":
                collect_command("apt packages", "apt list --installed > packages-apt.txt")
                artifact.collect("/etc/os-release", "os-release.txt")
            artifact.collect("*.txt")

        with t.cwd(log.logpath):
            artifact.collect("*.log", "logs/")

        acache.commit(artifact)
        if acache.upload_enabled():
            assert acache.upload(artifact)
            url = acache.location(artifact)
        else:
            url = f"{artifact.identity}.tgz"
            t.archive(artifact.path, url)

        log.info("Location: {}", url)



from configparser import ConfigParser, NoOptionError, NoSectionError
from urllib.parse import urlparse
import os
import re

from jolt import common_pb2 as common_pb
from jolt import filesystem as fs
from jolt import utils
from jolt.error import raise_error_if
from jolt.manifest import ManifestExtension, ManifestExtensionRegistry


_workdir = os.getcwd()


if os.getenv("JOLT_CONFIG_PATH"):
    location = fs.path.join(os.getenv("JOLT_CONFIG_PATH"), "config")
    location_user = fs.path.join(os.getenv("JOLT_CONFIG_PATH"), "user")
elif os.name == "nt":
    appdata = os.getenv("APPDATA", fs.path.join(fs.userhome(), "AppData", "Roaming"))
    location = fs.path.join(appdata, "Jolt", "config")
    location_user = fs.path.join(appdata, "Jolt", "user")
else:
    location = fs.path.join(fs.userhome(), ".config", "jolt", "config")
    location_user = fs.path.join(fs.userhome(), ".config", "jolt", "user")


class ConfigFile(ConfigParser):
    def __init__(self, location, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._location = location
        if location is not None:
            dirname = fs.path.dirname(location)
            if dirname:
                fs.makedirs(dirname)

    def load(self):
        if self._location:
            super().read(self._location)
            if not self.has_section("jolt"):
                self.add_section("jolt")

    def save(self, path=None):
        if self._location is None and path is None:
            return
        with open(path or self._location, 'w') as configfile:
            super().write(configfile)

    def delete(self, section, key):
        if key is None:
            return self.remove_section(section)
        try:
            success = self.remove_option(section, key)
            if success and len(self[section].items()) <= 0:
                self.remove_section(section)
            return success
        except NoSectionError:
            return False

    def set(self, section, key, value):
        if not self.has_section(section):
            self.add_section(section)
        super().set(section, key, value)


class Config(object):
    def __init__(self):
        self._configs = []

    def configs(self, alias=None):
        return [config for name, config in self._configs if not alias or name == alias]

    def add_file(self, alias, location):
        file = ConfigFile(location)
        if alias == "cli":
            self._configs.insert(-1, (alias, file))
        else:
            self._configs.append((alias, file))
        return file

    def get(self, section, key, default, alias=None):
        for config in reversed(self.configs(alias)):
            try:
                return config.get(section, key)
            except (NoOptionError, NoSectionError):
                continue
        return default

    def set(self, section, key, value, alias=None):
        count = 0
        for config in self.configs(alias):
            config.set(section, key, value)
            count += 1
        return count

    def delete(self, section, key, alias=None):
        count = 0
        for config in self.configs(alias):
            count += int(config.delete(section, key))
        return count

    def sections(self, alias=None):
        s = []
        for config in self.configs(alias):
            s += config.sections()
        return sorted(s)

    def options(self, section, alias=None):
        s = []
        for config in self.configs(alias):
            if config.has_section(section):
                s += config[section].items()
        return sorted(s)

    def items(self, alias=None):
        o = {}
        for section in self.sections(alias):
            for option, value in self.options(section):
                o[(section, option)] = value
            if not o:
                o[(section, None)] = None
        return [(section, option, value) for (section, option), value in o.items()]

    def load(self):
        for config in self.configs():
            config.load()

    def save(self, path=None):
        for name, config in self._configs:
            config.save(os.path.join(path, f"{name}.conf") if path else None)


_config = Config()
_config.add_file("global", location)
_config.add_file("user", location_user)
_manifest = _config.add_file("manifest", None)
# Note: cli configs are added next to last in the chain,
# before manifest, and can therefore not override the
# imported manifest config.
_config.add_file("cli", None)
_config.load()


def get(section, key, default=None, expand=True, alias=None):
    val = _config.get(section, key, default, alias)
    return utils.expand(val) if expand and val is not None else val


def getint(section, key, default=None, alias=None):
    return int(get(section, key, default=default, alias=alias))


def getsize(section, key, default=None, alias=None):
    units = {
        None: 1,
        "B": 1,

        "K": 1000,
        "M": 1000**2,
        "G": 1000**3,
        "T": 1000**4,
        "P": 1000**5,
        "E": 1000**6,

        "Ki": 1024,
        "Mi": 1024**2,
        "Gi": 1024**3,
        "Ti": 1024**4,
        "Pi": 1024**5,
        "Ei": 1024**6,
    }

    value = get(section, key, default=None, alias=alias)
    if value is None:
        if type(default) is int:
            return default
        else:
            value = str(default)

    m = re.search(r"^(0|[1-9][0-9]*) ?([KMGTPE]i?)?B?$", value)
    raise_error_if(
        not m,
        "Config: size '{0}' invalid for '{1}.{2}', expected '<size> <unit>'", value, section, key)
    raise_error_if(
        m[2] not in units,
        "Config: unit invalid for '{0}.{1}', expected [B,K,M,G,T,P,E,Mi,Gi,Ti,Pi,Ei]", section, key)

    return int(m[1]) * units.get(m[2], 1)


def getfloat(section, key, default=None, alias=None):
    return float(get(section, key, default=default, alias=alias))


def getboolean(section, key, default=None, alias=None):
    value = get(section, key, default=default, alias=alias)
    return value is not None and str(value).lower() in ["true", "yes", "on", "1"]


def geturi(section, key, default=None, alias=None):
    value = get(section, key, default=default, alias=alias)
    return urlparse(value)


def get_jolthome():
    if os.name == "nt":
        return fs.path.join(os.getenv("LOCALAPPDATA", fs.path.join(fs.userhome(), "AppData", "Local")), "Jolt")
    else:
        return fs.path.join(fs.userhome(), ".jolt")


def get_logpath():
    return get("jolt", "logpath", get_jolthome())


def get_cachedir():
    if os.name == "nt":
        default = fs.path.join(get_jolthome(), "Cache")
    else:
        default = fs.path.join(fs.userhome(), ".cache", "jolt")
    return get("jolt", "cachedir", default)


def get_workdir():
    return _workdir


def get_shell():
    cmd = os.getenv("SHELL", "bash")
    if "bash" in cmd:
        cmd = cmd + " --norc"
    return get("jolt", "shell", cmd)


def set(section, key, value, alias=None):
    _config.set(section, key, value, alias or "user")


def load(file):
    _config.load()


def load_or_set(file_or_str):
    if fs.path.exists(file_or_str):
        _config.add_file("cli", file_or_str)
        _config.load()
    else:
        key_value = file_or_str.split("=", 1)
        raise_error_if(len(key_value) <= 1, "Syntax error in configuration: '{}'".format(file_or_str))
        section_key = key_value[0].split(".", 1)
        raise_error_if(len(section_key) <= 1, "Syntax error in configuration: '{}'".format(file_or_str))
        _config.set(section_key[0], section_key[1], key_value[1], alias="cli")


def save(path=None):
    _config.save(path)


def delete(key, alias=None):
    section, option = split(key)
    return _config.delete(section, option, alias)


def sections(alias=None):
    return _config.sections(alias)


def plugins():
    automatic = ["cxxinfo", "environ", "paths", "python", "strings"]
    return sections() + automatic


def items(alias=None):
    return _config.items(alias)


def options(section, alias=None):
    return _config.options(section, alias)


def split(string):
    try:
        section, key = string.split(".", 1)
    except ValueError:
        section, key = string, None
    return section, key


class ConfigExtension(ManifestExtension):
    def export_manifest(self, manifest, _):
        manifest.config = get("network", "config", "", expand=False)

        for key, value in options("params"):
            p = manifest.create_parameter()
            p.key = "config." + key
            p.value = value

    def import_manifest(self, manifest):
        if manifest.config:
            _manifest.read_string(manifest.config)
            from jolt.loader import JoltLoader
            JoltLoader.get().load_plugins()

        for param in manifest.parameters:
            if param.key.startswith("config."):
                set("params", param.key.split(".", 1)[1], param.value)

    def import_protobuf(self, pb):
        self.import_manifest(pb)


def export_config():
    return get("network", "config", "", expand=False)


def export_params():
    parameters = []
    for key, value in options("params"):
        parameters.append(common_pb.Property(key="config." + key, value=value))
    return parameters


# High priority so that plugins are loaded before resources are acquired.
ManifestExtensionRegistry.add(ConfigExtension(), -10)



from requests.exceptions import ConnectTimeout, RequestException
from urllib.parse import urlparse, urlunparse

from jolt import utils
from jolt import cache
from jolt import log
from jolt import config
from jolt import filesystem as fs
from jolt import tools
from jolt.error import raise_error_if, JoltError


NAME = "cache"
NAME_LOG = "Cache Service"
TIMEOUT = (3.5, 27)
TIMEOUT_HEAD = (27, 27)


class Cache(cache.StorageProvider):
    def __init__(self, cache):
        super().__init__()
        self._cache = cache
        self._uri = config.get(NAME, "uri", "http://cache")
        self._uri = self._uri.rstrip("/")
        raise_error_if(not self._uri, "Cache Service URI not configured")
        self._file_uri = self._uri + "/files"
        self._upload = config.getboolean(NAME, "upload", True)
        self._download = config.getboolean(NAME, "download", True)
        self._disabled = False

    def _get_path(self, artifact):
        return artifact.tools.expand(
            "{name}/{file}",
            name=artifact.task.name,
            file=fs.path.basename(artifact.get_archive_path()))

    def _get_url(self, artifact):
        return artifact.tools.expand(
            "{uri}/{name}/{file}",
            uri=self._file_uri,
            name=artifact.task.name,
            file=fs.path.basename(artifact.get_archive_path()))

    @utils.retried.on_exception((RequestException, JoltError))
    def download(self, artifact, force=False):
        if self._disabled:
            return False
        if not self._download and not force:
            return False
        url = self._get_url(artifact)
        return artifact.tools.download(
            url,
            artifact.get_archive_path(),
            exceptions=False,
            timeout=TIMEOUT)

    def download_enabled(self):
        return not self._disabled and self._download

    @utils.retried.on_exception((RequestException))
    def upload(self, artifact, force=False):
        if self._disabled:
            return True
        if not self._upload and not force:
            return True
        url = self._get_url(artifact)
        archive = artifact.get_archive()
        return artifact.tools.upload(
            archive, url,
            exceptions=False,
            timeout=TIMEOUT)

    def upload_enabled(self):
        return not self._disabled and self._upload

    @utils.retried.on_exception((RequestException))
    def availability(self, artifacts):
        if self._disabled:
            return [], artifacts

        file_map = {self._get_path(artifact): artifact for artifact in artifacts}
        try:
            data = {"files": list(file_map.keys())}
            response = tools.http_session.post(self._file_uri, json=data, stream=True, timeout=TIMEOUT_HEAD)
        except ConnectTimeout:
            self._disabled = True
            log.warning(NAME_LOG + " failed to establish server connection, disabled")
            return []

        if response.status_code != 200:
            log.debug(NAME_LOG + " POST ({}): {}", response.status_code, self._file_uri)
            return [], artifacts

        present = []
        missing = []
        data = response.json()

        for file in data.get("present", []):
            present.append(file_map[file])

        for file in data.get("missing", []):
            missing.append(file_map[file])

        return present, missing

    @utils.retried.on_exception((RequestException))
    def location(self, artifact):
        if self._disabled:
            return False

        url = self._get_url(artifact)
        try:
            response = tools.http_session.head(url, stream=True, timeout=TIMEOUT_HEAD)

        except ConnectTimeout:
            self._disabled = True
            log.warning(NAME_LOG + " failed to establish server connection, disabled")
            return False

        if response.status_code != 200:
            log.debug(NAME_LOG + " HEAD ({}): {}", response.status_code, url)
            return None

        url = urlunparse(urlparse(url))
        return url


@cache.RegisterStorage
class HttpFactory(cache.StorageProviderFactory):
    @staticmethod
    def create(cache):
        log.verbose(NAME_LOG + " Loaded")
        return Cache(cache)



"""
This module provides the :class:`GTestRunner` task base class, a derivation
of :class:`jolt.tasks.Runner` for running Google Test applications.
Several parameters can be assigned or overidden to control the behavior.

Example:

  .. literalinclude:: ../examples/googletest/runner.jolt
    :language: python
    :caption: examples/googletest/runner.jolt

The module also provides a set of task decorators for use with standalone
task classes. They set different ``GTEST_*`` environment variables to control
the behavior of Google Test executables in the same manner as the parameters in
:class:`GTestRunner` do.


Example:

  .. code-block:: python

    from jolt import Task
    from jolt.plugins import googletest

    # ....

    @googletest.fail_fast(default=True)
    @googletest.repeat(default=10)
    @googletest.filter()
    @googletest.junit_report()
    class TestRunner(Task):
        requires = ["test"]

        def run(self, deps, tools):
            tools.run("unittest")

  .. code-block:: bash

    $ jolt build testrunner:repeat=1,filter=TestModule.*

"""

from functools import wraps
from os import path
import json

from jolt import BooleanParameter, IntParameter, Parameter, Runner
from jolt import influence
from jolt.error import raise_task_error, raise_task_error_if
from jolt.plugins import junit
from jolt.utils import deprecated, ignore_exception


@deprecated
def import_failures(xml, report):
    return junit.import_junit_report(xml, report, errors=True, failures=True)


def break_on_failure(default: bool = False, param: bool = True, attr: str = "break_on_failure"):
    """
    Task class decorator controlling the GTEST_BREAK_ON_FAILURE environment variable.

    When the variable is set, test applications failures trigger breakpoints. On Linux,
    this typically results in a core dump that can be loaded into the debugger.

    The value is taken from the parameter ``break_on_failure`` which is created
    by default. If no parameter is created, the value is instead taken from the class
    attribute with the same name. The name of the paramter/attribute can be changed.

    Args:
        default (boolean): Default value of the parameter and env. variable.
            Default: False.
        param (boolean): Create a task parameter. Default: True.
        attr (str): Name of parameter or class attribute.

    """

    def decorator(cls):
        _old_run = cls.run

        @wraps(cls.run)
        def run(self, deps, tools):
            with tools.environ(GTEST_BREAK_ON_FAILURE=str(int(bool(getattr(self, attr, default))))):
                _old_run(self, deps, tools)

        if param:
            break_on_failure_param = BooleanParameter(default, help="Trigger breakpoint on failures.")
            break_on_failure_param.__set_name__(cls, attr)
            setattr(cls, attr, break_on_failure_param)

        cls.run = run
        return influence.source("googletest.break_on_failure", break_on_failure)(cls)

    return decorator


def brief(default: bool = False, param: bool = True, attr: str = "brief"):
    """
    Task class decorator controlling the GTEST_BRIEF environment variable.

    When the variable is set, only failed test-cases are logged.

    The value is taken from the parameter ``brief`` which is created
    by default. If no parameter is created, the value is instead taken from the class
    attribute with the same name. The name of the paramter/attribute can be changed.

    GoogleTest version >= 1.11 is required.

    Args:
        default (boolean): Default value of the parameter and env. variable.
            Default: False.
        param (boolean): Create a task parameter. Default: True.
        attr (str): Name of parameter or class attribute.

    """

    def decorator(cls):
        _old_run = cls.run

        @wraps(cls.run)
        def run(self, deps, tools):
            with tools.environ(GTEST_BRIEF=str(int(bool(getattr(self, attr, default))))):
                _old_run(self, deps, tools)

        if param:
            brief_param = BooleanParameter(default, help="Only print failed test-cases.")
            brief_param.__set_name__(cls, attr)
            setattr(cls, attr, brief_param)

        cls.run = run
        return influence.source("googletest.brief", brief)(cls)

    return decorator


def disabled(default: bool = False, param: bool = True, attr: str = "disabled"):
    """
    Task class decorator controlling the GTEST_ALSO_RUN_DISABLED_TESTS environment variable.

    When the variable is set, disabled test-cases are also run.

    The value is taken from the parameter ``disabled`` which is created
    by default. If no parameter is created, the value is instead taken from the class
    attribute with the same name. The name of the paramter/attribute can be changed.

    Args:
        default (boolean): Default value of the parameter and env. variable.
            Default: False.
        param (boolean): Create a task parameter. Default: True.
        attr (str): Name of parameter or class attribute.

    """

    def decorator(cls):
        _old_run = cls.run

        @wraps(cls.run)
        def run(self, deps, tools):
            with tools.environ(GTEST_ALSO_RUN_DISABLED_TESTS=str(int(bool(getattr(self, attr, default))))):
                _old_run(self, deps, tools)

        if param:
            disabled_param = BooleanParameter(default, help="Also run disabled test-cases.")
            disabled_param.__set_name__(cls, attr)
            setattr(cls, attr, disabled_param)

        cls.run = run
        return influence.source("googletest.disabled", disabled)(cls)

    return decorator


def fail_fast(default: bool = False, param: bool = True, attr: str = "fail_fast"):
    """
    Task class decorator controlling the GTEST_FAIL_FAST environment variable.

    When the variable is set, test applications will abort when the first failure is found.

    The value is taken from the parameter ``fail_fast`` which is created
    by default. If no parameter is created, the value is instead taken from the class
    attribute with the same name. The name of the paramter/attribute can be changed.

    Args:
        default (boolean): Default value of the parameter and env. variable.
            Default: False.
        param (boolean): Create a task parameter. Default: True.
        attr (str): Name of parameter or class attribute.

    """

    def decorator(cls):
        _old_run = cls.run

        @wraps(cls.run)
        def run(self, deps, tools):
            with tools.environ(GTEST_FAIL_FAST=str(int(bool(getattr(self, attr, default))))):
                _old_run(self, deps, tools)

        if param:
            fail_fast_param = BooleanParameter(default, help="Stop when the first failure is found.")
            fail_fast_param.__set_name__(cls, attr)
            setattr(cls, attr, fail_fast_param)

        cls.run = run
        return influence.source("googletest.fail_fast", fail_fast)(cls)

    return decorator


def filter(default: str = "*", param: bool = True, attr: str = "filter"):
    """
    Task class decorator controlling the GTEST_FILTER environment variable.

    The variable instructs test applications to only run test-cases that matches
    a wildcard filter string (default: *).

    The value is taken from the parameter ``filter`` which is created
    by default. If no parameter is created, the value is instead taken from the class
    attribute with the same name. The name of the paramter/attribute can be changed.

    Args:
        default (boolean): Default value of the parameter and env. variable.
            Default: False.
        param (boolean): Create a task parameter. Default: True.
        attr (str): Name of parameter or class attribute.

    """

    def decorator(cls):
        _old_run = cls.run

        @wraps(cls.run)
        def run(self, deps, tools):
            with tools.environ(GTEST_FILTER=str(getattr(self, attr, default))):
                _old_run(self, deps, tools)

        if param:
            filter_param = Parameter(default, help="Test-case filter")
            filter_param.__set_name__(cls, attr)
            setattr(cls, attr, filter_param)

        cls.run = run
        return influence.source("googletest.filter", filter)(cls)

    return decorator


def junit_report():
    """
    Decorator enabling JUnit test reporting in Google Test applications.

    The decorator sets the GTEST_OUTPUT environment, thereby instructing
    test applications to write a JUnit report file upon test completion.
    The task publishes the report in the task artifact under report/junit/.

    Any errors found in the report are parsed and attached to the task.
    They are included in email reports if the email plugin is enabled.
    """

    def decorator(cls):
        _old_run = cls.run
        _old_publish = cls.publish

        @wraps(cls.run)
        def run(self, deps, tools):
            raise_task_error_if(
                tools.getenv("GTEST_OUTPUT"),
                self,
                "GoogleTest output already enabled, cannot proceed with JUnit report setup")
            gtestdir = tools.builddir("gtest-junit-report")
            gtestreport = path.join(gtestdir, "report.xml")
            with tools.environ(GTEST_OUTPUT="xml:" + gtestreport):
                try:
                    _old_run(self, deps, tools)
                finally:
                    with ignore_exception(), self.report() as report:
                        junit.import_junit_report(gtestreport, report)

        @wraps(cls.publish)
        def publish(self, artifact, tools):
            _old_publish(self, artifact, tools)
            gtestdir = tools.builddir("gtest-junit-report")
            with tools.cwd(gtestdir):
                artifact.collect("*", "report/junit/")

        cls.run = run
        cls.publish = publish
        return influence.source("googletest.junit_report", junit_report)(cls)

    return decorator


def json_report():
    """
    Decorator enabling JSON test reporting in Google Test applications.

    The decorator sets the GTEST_OUTPUT environment, thereby instructing
    test applications to write a JSON report file upon test completion.
    The task publishes the report in the task artifact under report/json/.

    Any errors found in the report are parsed and attached to the task.
    They are included in email reports if the email plugin is enabled.
    """

    def decorator(cls):
        _old_run = cls.run
        _old_publish = cls.publish

        @wraps(cls.run)
        def run(self, deps, tools):
            raise_task_error_if(
                tools.getenv("GTEST_OUTPUT"),
                self,
                "GoogleTest output already enabled, cannot proceed with JSON report setup")
            gtestdir = tools.builddir("gtest-json-report")
            gtestreport = path.join(gtestdir, "report.json")
            with tools.environ(GTEST_OUTPUT="json:" + gtestreport):
                try:
                    _old_run(self, deps, tools)
                finally:
                    with ignore_exception(), self.report() as report:
                        with open(gtestreport) as fp:
                            json_report = json.load(fp)
                        for testsuite in json_report["testsuites"]:
                            for testcase in testsuite["testsuite"]:
                                failures = testcase.get("failures", [])
                                for failure in failures:
                                    name = f"{testsuite['name']}.{testcase['name']}"
                                    message = failure["failure"]
                                    report.add_error("Test Failed", name, message)

        @wraps(cls.publish)
        def publish(self, artifact, tools):
            _old_publish(self, artifact, tools)
            gtestdir = tools.builddir("gtest-json-report")
            with tools.cwd(gtestdir):
                artifact.collect("*", "report/json/")

        cls.run = run
        cls.publish = publish
        return influence.source("googletest.json_report", json_report)(cls)

    return decorator


def repeat(default: int = 1, param: bool = True, attr: str = "repeat"):
    """
    Task class decorator controlling the GTEST_REPEAT environment variable.

    The variable instructs test applications to repeat test-cases the specified
    number of times.

    The value is taken from the parameter ``repeat`` which is created
    by default. If no parameter is created, the value is instead taken from the class
    attribute with the same name. The name of the paramter/attribute can be changed.

    Args:
        default (boolean): Default value of the parameter and env. variable.
            Default: False.
        param (boolean): Create a task parameter. Default: True.
        attr (str): Name of parameter or class attribute.

    """

    def decorator(cls):
        _old_run = cls.run

        @wraps(cls.run)
        def run(self, deps, tools):
            try:
                repeat = int(getattr(self, attr, default))
            except ValueError:
                raise_task_error(self, f"Value assigned to '{attr}' is not an integer")
            with tools.environ(GTEST_REPEAT=str(repeat)):
                _old_run(self, deps, tools)

        if param:
            repeat_param = IntParameter(default, help="Test-case repetitions.")
            repeat_param.__set_name__(cls, attr)
            setattr(cls, attr, repeat_param)

        cls.run = run
        return influence.source("googletest.repeat", repeat)(cls)

    return decorator


def seed(default: int = 0, param: bool = True, attr: str = "seed"):
    """
    Task class decorator controlling the GTEST_RANDOM_SEED environment variable.

    The variable sets an initial value for the random number generator used to
    shuffle test-cases. In order to reproduce ordering issues, a user may assign
    a specific seed to get the same test-case order. The default value is 0
    which causes Google Test to use time as the seed.

    The value is taken from the parameter ``seed`` which is created
    by default. If no parameter is created, the value is instead taken from the class
    attribute with the same name. The name of the paramter/attribute can be changed.

    Args:
        default (boolean): Default value of the parameter and env. variable.
            Default: False.
        param (boolean): Create a task parameter. Default: True.
        attr (str): Name of parameter or class attribute.

    """

    def decorator(cls):
        _old_run = cls.run

        @wraps(cls.run)
        def run(self, deps, tools):
            try:
                seed = int(getattr(self, attr, default))
            except ValueError:
                raise_task_error(self, f"Value assigned to '{attr}' is not an integer")
            with tools.environ(GTEST_RANDOM_SEED=str(seed)):
                _old_run(self, deps, tools)

        if param:
            seed_param = IntParameter(default, min=0, max=99999, help="Random number generator seed.")
            seed_param.__set_name__(cls, attr)
            setattr(cls, attr, seed_param)

        cls.run = run
        return influence.source("googletest.seed", seed)(cls)

    return decorator


def shuffle(default: bool = False, param: bool = True, attr: str = "shuffle"):
    """
    Task class decorator controlling the GTEST_SHUFFLE environment variable.

    When set, the test application runs its test-cases in random order.
    In order to reproduce test-case ordering issues, a user may assign a specific
    random number generator seed to get the same test-case execution order.
    See :func:`seed`.

    The value is taken from the parameter ``shuffle`` which is created
    by default. If no parameter is created, the value is instead taken from the class
    attribute with the same name. The name of the paramter/attribute can be changed.

    Args:
        default (boolean): Default value of the parameter and env. variable.
            Default: False.
        param (boolean): Create a task parameter. Default: True.
        attr (str): Name of parameter or class attribute.

    """

    def decorator(cls):
        _old_run = cls.run

        @wraps(cls.run)
        def run(self, deps, tools):
            with tools.environ(GTEST_SHUFFLE=str(int(bool(getattr(self, attr, default))))):
                _old_run(self, deps, tools)

        if param:
            shuffle_param = BooleanParameter(default, help="Randomize test-case execution order.")
            shuffle_param.__set_name__(cls, attr)
            setattr(cls, attr, shuffle_param)

        cls.run = run
        return influence.source("googletest.shuffle", shuffle)(cls)

    return decorator


@break_on_failure()
@brief()
@disabled()
@fail_fast()
@filter()
@junit_report()
@repeat()
@seed()
@shuffle()
class GTestRunner(Runner):
    """
    Abstract task base class that runs Google Test applications.

    Test executables are consumed from requirement artifacts. The
    artifacts must export metadata indicating the name of the binary,
    using ``artifact.strings.executable``.

    A number of parameters can be assigned from the command line
    to control the behavior of the task. Parameters can also be
    overridden using regular Python class attributes in subclasses.

    A JUnit test report is generated and published in the task artifact.
    The results of the report are also imported into Jolt and distributed
    with report emails if the email plugin has been enabled.

    Example:

    .. literalinclude:: ../examples/googletest/runner.jolt
      :language: python
      :caption: examples/googletest/runner.jolt
    """

    abstract = True

    break_on_failure = BooleanParameter(False, help="Trigger breakpoint on failures.")
    """ Trigger breakpoint on test-case failure. """

    brief = BooleanParameter(False, help="Only print failed test-cases.")
    """ Only print failed test-cases. """

    disabled = BooleanParameter(False, help="Also run disabled test-cases.")
    """ Also run disabled test-cases. """

    fail_fast = BooleanParameter(False, help="Stop when the first failure is found.")
    """ Stop when the first failure is found. """

    filter = Parameter("*", help="Test-case filter.")
    """ Test-case filter. """

    repeat = IntParameter(1, help="Test-case repetitions.")
    """ Test-case repetitions. """

    seed = IntParameter(0, min=0, max=99999, help="Random number generator seed.")
    """ Random number generator seed. """

    shuffle = BooleanParameter(True, help="Randomize test-case execution order.")
    """ Randomize test-case execution order. """
